//
//  Base classes of GUI gadgets.
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <guit.hpp>
#include <giobuffer.hpp>
#include <ggraphics.hpp>
#include <gcomputedstyle.hpp>
#include <gvalue.hpp>
#include <gtext.hpp>
#include <gfloat.hpp>
#include <gint.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
#include <gtextrender.hpp>
#include <gstyle.hpp>
#include <gadgetstyle.hpp>
using namespace std;
namespace guit {

// for helping selection of very small shapes
#define SmallShape 2.f

static const GString FileSignature = "#guit gadgets v1";

GString const Gadget::emptystring_;

GClone operator|(GClone a, GClone b) {
  return GClone(uint8_t(a) | uint8_t(b));
}

uint8_t operator&(GClone a, GClone b) {
  return uint8_t(a) & uint8_t(b);
}

// - - - GadgetType - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

// pourrait utiliser IDs !!!
static std::unordered_map<GString,GadgetType*>& getClasses() {
  static std::unordered_map<GString,GadgetType*> classes;
  return classes;
}

GadgetType* GadgetType::findType(GString const& name) {
  auto it = getClasses().find(name);
  if (it != getClasses().end()) return it->second;
  else return nullptr;
}

GadgetType::GadgetType(GString const& name) : name_(name) {
  id_ = GCore::obtainId(name_);

  if (findType(name)) {
    gerror("GadgetType",
           "This class (" + name_ + ") already exists (not added to class base)");
  }
  else {
    //getClasses().insert_or_assign(name_, this);
    getClasses()[name_] = this;
  }

  style_ = GStyleBase::instance().findStyle(id_);
  if (style_) style_->realize();
  else throw GFatalError("GadgetType",
                         "Fatal error: style of '"+name_+"' gadget class not found");
}

Gadget* GadgetType::createInstance(GString const& cname, GString const& arg) {
  auto* cl = findType(cname);
  return cl ? &cl->newInstance(arg) : nullptr;
}

void GadgetType::operator delete(void* ptr) {
  gerror("GadgetType::delete",
         "the delete operator can't be used to destroy a GadgetType (ignored)");
}

bool GadgetType::operator==(GadgetType const& other) const {
  return this == &other;
}

bool GadgetType::operator!=(GadgetType const& other) const {
  return this != &other;
}

// - - - Gadget - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

struct GadgetType__ : public GadgetType {
  GadgetType__() : GadgetType("Gadget") {}

  Gadget& newInstance(GString const& title) {
    throw GFatalError("GadgetType","cant instantiate Gadget class");
  }
};

GadgetType& Gadget::Type() {
  static auto& t = *new GadgetType__;
  return t;
}

Gadget::Gadget(GadgetType& t) {
  modes_ = t.style_->modes_;
}

Gadget::Gadget(GadgetType& t, GString const& spec) {
  modes_ = t.style_->modes_;
  
  if (!spec.empty()) {
    GString title;
    GadgetStyle::initStyle(*this, spec, title);
    if (!title.empty()) add(title);
  }
}

Gadget::~Gadget() {
  // cerr << "Gadget::~Gadget() "<< this << " " << typeName() << endl;
  if (parent_) {
    if (auto* par = getParent()) {
      // this object is going to be deleted anyway, so it makes
      // no sense auto-deleting it
      ignoreSmartPointers();
      par->remove(this);
    }
  }
  //suffit pas: tous les props mutables qui pointent dessus !!
  //devraient etre reiinit

  // removes props and owners of objects
  // not removeProps() because it call update(), which is useless and buggy
  props_.removeAll(this);
  delete shape_;
}

void Gadget::initClone(Gadget& from, GClone const& c) {
  style_ = from.style_;

  if (c & GClone::ShareProps) {
    modes_ = from.modes_;
    for (auto& p : from.props_) {
      if (!p->toNotifier()) {    // notifiers never shared!
        *this << *p;
      }
    }
  }
  else if (c & GClone::Props) {
    modes_ = from.modes_;
    for (auto& p : from.props_) {
      // not a notifier nor a condition nor a value
      if (!p->toNotifier() && !p->toCond() && p.get() != from.valueProp()) {
        if (auto* prop = p->clone(true)) *this << prop;
      }
    }
    // needed because bindings will be copied although p.get() != from.valueProp()
    clearValues();
  }
}


void Gadget::init(GRender* parrender) {
  if (omodes_.initialized) return;

  if (style_) style_->realize(*this);
  omodes_.initialized = 1;
  GRender r(*this, parrender);

  // if the gadget has a value, search the value in its style
  if (style_ && hasValue()) {
    if (auto* p = style_->findProp<GValue>(*this)) setValue(p->value_);
  }

  for (auto& p : props_) {
    if (auto* m = p->toMutableProp()) {
      // mettre a jour les props mutables qui sont definies dans les styles
      // note: n'inclut pas props heritees ou des styles contextuels !!
      if (style_ && !m->pmodes_.isSet) {
        if (auto* sp = style_->findProp(*this, m->type())) m->set(*sp);
      }
    }
    else if (auto* h = p->to<GadgetHolder>()) {
      if (h->gadget_) h->gadget_->init(nullptr);
    }
  }

  if (auto* box = toBox()) {
    box->initChildren(&r);
  }

  if ((eventmask_ & GEventType::Init) != 0) {
    GEvent e(on::init, GCore::elapsedTime());
    doFire(e);
  }
}

// - - - Types, Modes, Parent, Window - - - - - - - - - - - - - - - - - - - - -

GWindow* Gadget::getWindow() const {
  if (GWindow* win = const_cast<Gadget*>(this)->toWindow()) return win;

  Gadget* p = parent_;
  while (p) {
    if (p->toWindow()) return p->toWindow();
    // !!!BUG qq part, evite infinite loop !!!
    if (p->parent_ && p->parent_->parent_ == p) return nullptr;
    p = p->parent_;
  }
  return nullptr;
}

Gadget& Gadget::setUnclippedLayers(bool val) {
  modes_.unclippedLayers = val;
  return *this;
}

Gadget& Gadget::catchEvents(bool val) {
  modes_.catchEvents = val;
  return *this;
}

Gadget& Gadget::ignoreEvents(bool val) {
  modes_.ignoreEvents = val;
  return *this;
}

Gadget& Gadget::setActivable(bool s) {
  modes_.activable = s;
  return *this;
}

Gadget& Gadget::setDownActivable(bool s) {
  modes_.downActivable = s;
  if (s) modes_.activable = true;
  return *this;
}

Gadget& Gadget::setCheckable(bool s) {
  modes_.checkable = s;
  if (s) modes_.activable = true;
  return *this;
}

// - - - Spec and Style - - - - - - - - - - - - - - - - - - - - - - - - - - -

class GadgetOpener : public GadgetHolder {
public:
  // WARNING: g g must be a menu or a window!
  GadgetOpener(GTrigger& t, Gadget& g, int button, bool autoplace)
  : GadgetHolder(g), trigger_(t), button_(button), autoplace_(autoplace) {}

  static GPropType& Type();
  GPropType& type() const override {return Type();}

  bool onAdd(Gadget* g) override {
    if (g) g->eventmask_ = g->eventmask_ | trigger_.eventMask();
    return true;
  }

  void onRemove(Gadget*) override {
    //if (g)  update eventmask !!!!                   // TODO !!!!
  }

  void fire(GEvent& e) override {
    if (e.matches(trigger_,-1) > 0) {
      if (auto* m = gadget_->toMenu()) m->show(e, autoplace_);
      else if (auto* w = gadget_->toWindow()) w->show(e, autoplace_);
    }
  }

  GTrigger& trigger_;
  int button_{};
  bool autoplace_{};
};

GPropType& GadgetHolder::Type() {
  static auto& t = *new GPropType("GadgetHolder", GPropType::Holder);
  return t;
}

GPropType& GadgetOpener::Type() {
  static auto& t = *new GPropType("GadgetOpener", GPropType::Holder);
  return t;
}

void Gadget::addOpener(Gadget& opened, GTrigger& trigger, int btn, bool autoplace) {
  if (settings_.hasOpener) {
    settings_.hasMenu = 0;
    if (auto* p = findProp<GadgetOpener>()) remove(p);
  }
  settings_.hasOpener = 1;
  if (opened.toMenu()) settings_.hasMenu = 1;
  // addprop not add, otherwise infinite loop with add(GadgetHolder*) !
  addProp(*new GadgetOpener(trigger, opened, btn, autoplace), true);
}

// - - - - - -

GItemSpec const& operator<<(GItemSpec const& sp, GString const& str) {
  sp.props_.push_back(new GText(str));
  return sp;
}

GItemSpec const& operator<<(GItemSpec const& sp, GProp& prop) {
  sp.props_.push_back(prop);
  return sp;
}

GItemSpec const& operator<<(GItemSpec const& sp, gptr<GProp> prop) {
  if (prop) sp.props_.push_back(prop); else gerror("add","Can't add null prop");
  return sp;
}

GItemSpec const& operator<<(GItemSpec const& sp, GFunction const& fun) {
  sp.props_.push_back(on::action / fun);
  return sp;
}

GItemSpec const& operator<<(GItemSpec const& sp, GEventFunction const& fun) {
   sp.props_.push_back(on::action / fun);
  return sp;
}

GItemSpec const& operator<<(GItemSpec const& sp, GMenu& menu) {
   sp.props_.push_back(new GadgetHolder(menu));
  return sp;
}

GItemSpec const& operator<<(GItemSpec const& sp, gptr<GMenu> menu) {
  if (menu) sp.props_.push_back(new GadgetHolder(*menu));
  else gerror("add","Can't add null menu");
  return sp;
}

GItemSpec const& operator<<(GItemSpec const& sp, GWindow& win) {
   sp.props_.push_back(new GadgetHolder(win));
  return sp;
}

GItemSpec const& operator<<(GItemSpec const& sp, gptr<GWindow> win) {
  if (win) sp.props_.push_back(new GadgetHolder(*win));
  else gerror("add","Can't add null window");
  return sp;
}

// - - - - - -

void Gadget::add(GMenu* opened) {
  if (!opened) error("add","Can't add null menu");
  else addOpener(*opened, on::mousedown, 2, false); // mousedown, button 2, onplace
}

void Gadget::add(GWindow* opened) {
  if (!opened) error("add","Can't add null window");
  // addprop not add, otherwise infinite loop with add(GadgetHolder*) !
  else addProp(*new GadgetHolder(*opened), true);  // just a holder, not opened
}

void Gadget::addFun(GFunction const& fun) {
  addProp(on::action / fun, true);
}

void Gadget::addFun(GEventFunction const& fun) {
  addProp(on::action / fun, true);
}

void Gadget::addFun(GCaller& fun) {
  addProp(on::action / fun, true);
}

void Gadget::add(GPropList const& list) {
  for (auto& it : list) addProp(*it, true);
}

void Gadget::add(GString const& str) {
  /* deux possibilités :
   - soit on ajoute un GText normalement, il pourra etre modifie et mettre
     a jour son parent
   - soit on ajoute un GText immutable, ca prend moins de ressources mais
     on peut pas le modifier
   */
  // addProp(*new GText(str), true);
  // or:
  auto* t = new GText(str);
  t->setImmutable();
  props_.addReplace(*t, this);
  update();
}

void Gadget::add(GProp* prop) {
  if (!prop) error("add","Can't add null child");
  else addProp(*prop, true);
}

// - - - - - -

void Gadget::addProp2(GProp& prop, bool add_to_list) {
  if (add_to_list) props_.addReplace(prop, this);
  else prop.onAdd(this);
}

// add_to_list is false for cond props: actions are performed but the prop
// is not actually added to the proplist
void Gadget::addProp(GProp& prop, bool add_to_list) {
  switch (prop.role()) {

    case GPropType::Cond:
    case GPropType::Notifier:
      addProp2(prop, true);   // true => always add
      // NB: no need to update
      return;

    case GPropType::Anim:
      addProp2(prop, add_to_list);
      if (auto* a = prop.toAnim()) a->addPropsTo(*this);
      return;

    case GPropType::Expr:
    case GPropType::ExprWrapper: {
      GExprWrapper* w{};
      if (auto* e = prop.toExpr()) w = new GExprWrapper(*e);
      else if (!(w = prop.toExprWrapper())) return; // should not occur
      auto* t = findProp<GText>();
      if (!t || !t->isMutable()) *this << *(t = new GText);
      w->bindWith(*t);
      addProp2(*w, add_to_list);
      updateLayout();
    } return;

    case GPropType::Number:
    case GPropType::Bool:
    case GPropType::Choice:
      if (!prop.isMutable()) {
        addProp2(*new GText(prop.stringValue()), add_to_list); // set immutable!
      }
      else {
        addProp2(prop, add_to_list);
        auto& text = this->prop<GText>();
        if (auto* pi = prop.toInt()) gsync(text, *pi, -1);
        else if (auto* pf = prop.toFloat()) gsync(text, *pf, -1);
        else if (auto* pb = prop.toBool()) gsync(text, *pb, -1);
      }
      updateLayout();
      return;
      
    default:
      addProp2(prop, add_to_list);
      if (auto* m = prop.toMutableProp()) m->startNotifyGadget(*this);
      return;
  }
}

// - - - -

bool Gadget::remove(GProp* p) {
  if (!p) {
    error("remove","Can't remove null child");
    return false;
  }
  else {
    if (!removeProp(*p, true)) return false;
    update();
    return true;
  }
}

bool Gadget::remove(GProp& p) {
  if (!removeProp(p, true)) return false;
  update();
  return true;
}

bool Gadget::remove(GPropType& t) {
  if (!props_.remove(this, t)) return false;
  update();
  return true;
}

bool Gadget::remove(GTrigger& t) {
  for (auto& it : props_) {
    if (auto* c = it->toCond()) {
      if (&c->trigger() == &t) {remove(it); return true;}
    }
  }
  return false;
}

void Gadget::removeProps() {   // devrait pas enlever l'id !!!
  props_.removeAll(this);
  update();
}

void Gadget::cleanup() {
  removeProps();
}

bool Gadget::removeProp2(GProp& prop, bool rem_from_list) {
  if (rem_from_list) {
    return props_.remove(this, prop);
  }
  else {
    prop.onRemove(this);
    return true;
  }
}

bool Gadget::removeProp(GProp& prop, bool rem_from_list) {
  switch (prop.role()) {

    case GPropType::Cond:
    case GPropType::Notifier:
      return removeProp2(prop, true);   // true => always remove

    case GPropType::Anim:
      if (auto* a = prop.toAnim()) {
        a->stop();
        a->removePropsFrom(*this);
      }
      return removeProp2(prop, rem_from_list);

    default:
      return removeProp2(prop, rem_from_list);
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - -- - - - - - - - - - - - - - -

bool Gadget::contain(GProp* prop) const {
  for (auto& p : props_) {
    if (p.get() == prop) return true;
  }
  return false;
}

void Gadget::cssProps(GPropMap& props) {
  GComputedStyle style(*this);
  props = style.getProps();
}

GProp* Gadget::findProp(GPropType& t) const {
  return props_.findProp(t);
}

GProp& Gadget::prop(GPropType& t, bool compstyle) {
  // search in proplist
  if (GProp* p = props_.findProp(t)) {
    if (p->isMutable()) {
      return *p;
    }
    // if not mutable duplicate prop in copy value mode
    else if (GProp* cop = p->clone(true)) {
      *this << cop;
      return *cop;
    }
    // if not clonable, create prop
    else {
      GProp& newp = t.newInstance();
      *this << newp;
      return newp;
    }
  }

  // if not in proplist, search in styles (if compute_style is true)
  if (compstyle) {
    GComputedStyle style(*this);
    // gptr<> to release ressources if not used
    if (gptr<GProp> p = style.getProp(t)) {
      *this << *p;
      return *p;
    }
  }

  GProp& newp = t.newInstance();         // if prop not found, create it
  if (auto* m = newp.toMutableProp()) {  // if mutable, copy value in style if any
    if (style_) {
      if (GProp* p = style_->findProp(*this, t)) m->set(*p);
    }
  }
  *this << newp;
  return newp;
}

GString Gadget::text() const {
  return emptystring_;
}

Gadget& Gadget::setRichText(bool mode) {
  modes_.richText = mode;
  return *this;
}

void Gadget::clearValues() {
  if (hasValue()) setValue("");
}

// - - - - - - - - - - - - - - - - - - - - - - - -- - - - - - - - - - - - - - -
/*
std::ostream& operator<<(std::ostream& out, Gadget const& g) {
  g.write(out, false);
  return out;
}
*/

bool Gadget::write(GString const& filename, GIOMode mode) const {
  GOFile out(filename);
  if (!out) return false;
  return write(out, mode);
}

bool Gadget::write(std::ostream& out, GIOMode mode) const {
  if (!out) return false;
  out.imbue(locale("C"));

  if (mode == GIOMode::Auto) mode = GIOMode::Values;

  GIOBuffer buf(out);

  try {
    if (mode == GIOMode::Values) {
      buf.setValueMode(true);
      writeValues(out, buf);
    }
    else {
      out << FileSignature << endl;
      buf.setValueMode(false);
      buf.skipbase_ = true;
      writeAll(out, buf);
    }
  }
  catch (GIOBuffer::IOError) {
    error("write", "error while writing file");
    return false;
  }

  return true;
}

void Gadget::writeValues(std::ostream& out, GIOBuffer& buf) const {
  if (isInternal()) return;

  if (hasValue()) {
    buf.printTabbed(stringValue());
    buf.printDelim();
  }
  auto* box = toBox();
  if (box && box->childCount() > 0) {
    buf.incrLevel();
    for (auto& it : box->children()) it->writeValues(out, buf);
    buf.decrLevel();
  }
}

void Gadget::writeAll(std::ostream& out, GIOBuffer& buf) const {
  if (isInternal()) return;

  if (!buf.skipbase_) {
    buf.printTabbed("- !");
    out << this->typeName() << "\n";
    buf.incrLevel();

    if (!props_.empty()) {
      bool found{};
      for (auto& p : props_) {
        if (p->isMutable()) {
          if (!found) {
            found = true;
            buf.printTabbed("props: {");
          }
          else {
            out << "; ";    // should be ,
          }
          out << p->typeName() << ": ";
          p->write(out);
        }
      }
      if (found) out << "}\n";
    }
    writeAux(out, buf);
  }

  auto* box = toBox();
  if (box && box->childCount() > 0) {
    if (buf.skipbase_) buf.skipbase_ = false;
    else {
      buf.printTabbed("children:\n");
      buf.incrLevel();
    }
    for (auto& it : box->children()) {it->writeAll(out, buf);}
    buf.decrLevel();
  }

  buf.decrLevel();
}

// - - - - - - - -
/*
std::istream& operator>>(std::istream& in, Gadget& g) {
  g.read(in, false);
  return in;
}*/

bool Gadget::read(GString const& filename, GIOMode mode) {
  GIFile in(filename);
  if (!in) return false;
  return read(in, mode);
}

bool Gadget::read(std::istream& in, GIOMode mode) {
  if (!in) return false;
  in.imbue(locale("C"));

  GIOBuffer buf(in);

  char c = 0;
  if (!in.get(c)) {
    error("read", "empty or invalid file");
    return false;
  }

  if (c != '#') in.putback(c);
  else {
    in.putback(c);
    GString line;
    getline(in, line);
    if (mode == GIOMode::Auto && line == FileSignature) mode = GIOMode::Content;
    buf.clear();
    buf.read();
  }
  if (mode == GIOMode::Auto) mode = GIOMode::Values;

  try {
    if (mode == GIOMode::Values) {
      buf.setValueMode(true);
      clearValues();
      readValues(in, buf);
    }
    else {  // Chidren mode
      buf.setValueMode(false);
      if (auto* box = toBox()) box->removeChildren();
      readAll(in, buf);
    }
  }
  catch (GIOBuffer::IOError) {
    error("read", "error while reading file");
    updateLayout();
    return false;
  }

  updateLayout();
  return true;
}

void Gadget::readValues(std::istream& in, GIOBuffer& buf) {
  if (hasValue() && buf.read()) {
    if (buf.tabCount() != buf.level() && buf.level() != -1) return;
    setValue(buf.data());
    buf.clear();
  }

  if (auto* box = toBox()) {
    buf.incrLevel();
    int k = 0, count = box->childCount();

    while (true) {
      if (!buf.read() || buf.tabCount() < buf.level()) {
        buf.decrLevel();
        return;
      }
      // if item exists update it without destroying it
      else if (k < count) {
        box->childAt(k)
        ->readValues(in, buf);
        k++;
      }
      // if item does not exist duplicate last line (if any)
      else if (auto* last = box->childAt(-1)) {
        auto& c = last->clone(GClone::Children | GClone::Props);
        *box << c;
        c.readValues(in, buf);
        k++;
      }
      else {    // BUG ??
        *box << "";   // creates default child
        auto* c = box->childAt(-1);
        c->readValues(in, buf);
        k++;
      }
    }
  }
}

void Gadget::readAll(std::istream& in, GIOBuffer& buf) {
  buf.incrLevel();

  // read props
  if (!buf.read() || buf.tabCount() < buf.level()) {
    buf.decrLevel();
    return;
  }
  if (buf.tabCount() > buf.level()) {
    error("read","invalid identation in: " + buf.data());
    throw GIOBuffer::IOError();
  }

  if (buf.data().compare(0, 8, "props: {") == 0) {
    GString props = buf.data().substr(8); 
    buf.clear();
    if (!GadgetStyle::parseProps(this, nullptr, props, ";}")) {
      error("read","invalid prop specification in: " + buf.data());
    }
  }

  // read aux
  if (!buf.read() || buf.tabCount() != buf.level()) {
    buf.decrLevel();
    return;
  }
  readAux(in, buf);

  auto* box = toBox();
  if (!box) {
    buf.decrLevel();
    return;
  }
  else {
    // read children
    if (!buf.read() || buf.tabCount() != buf.level()) {
      buf.decrLevel();
      return;
    }

    if (buf.data() == "children:") {
      buf.clear();
      buf.incrLevel();
    }
    else if (buf.level() != 0) {
      error("read", "wrong keyword in: " + buf.data());
      throw GIOBuffer::IOError();
      return;
    }

    while (true) {
      if (!buf.read() || buf.tabCount() != buf.level()) {
        buf.decrLevel();
        buf.decrLevel();
        return;
      }

      // read type
      if (buf.data().compare(0, 3, "- !") != 0) {
        error("read", "expecting - !type in:" + buf.data());
        throw GIOBuffer::IOError();
        return;
      }
      GString type = buf.data().substr(3);
      buf.clear();

      Gadget* child = GadgetType::createInstance(type, "");
      if (!child) {
        error("read", "unknown type " + type + " in: " + buf.data());
        throw GIOBuffer::IOError();
        return;
      }
      else {
        box->add(*child);
        child->readAll(in, buf);
      }
    }
  }
}

// - - - Callbacks and Behaviors  - - - - - - - - - - - - - - - - - - - - - - -

void Gadget::removeNotifiers(void* cond) {
  props_.removeNotifiers(cond);
}

void Gadget::doFire(GEvent& e) {
  if (props_.empty()) return;
  if ((eventmask_ & e.trigger().eventMask()) == 0) return;
  
  // avoir destructing object while firing callbacks
  gprotect<Gadget> protect(this);
  e.gadget_ = this;
  //GTrigger::setCurrentEvent(&e);
  //                    !!!! devrait gerer qu des elts peuvent etre enleves !!!
  //                    !!!! par les callbacks !!!
  for (auto& it : props_.list_) {
    it->fire(e);
  }
  //GTrigger::setCurrentEvent(nullptr);
}

Gadget& Gadget::enable(bool state) {
  if (state == modes_.enabled) return *this;
  modes_.enabled = state;
  settings_.trulyEnabled = settings_.trulyEnabled && modes_.enabled;

  if ((eventmask_ & GEventType::Enable) != 0) {
    if (state) {
      GEvent e(on::enabled, GCore::elapsedTime());
      doFire(e);
    }
    else {
      GEvent e(on::disabled, GCore::elapsedTime());
      doFire(e);
    }
  }
  update();
  return *this;
}

Gadget& Gadget::check(bool state) {
  // ne pas tester isSelectable(), incoherent et fait ailleurs
  selectImpl(state, Gadget::FromButton);
  return *this;
}

void Gadget::selectImpl(bool state, SelectFrom from) {
  if (state == modes_.checked) return;
  modes_.checkable = true;             // makes selectable silently!
  modes_.checked = state;

  if (from != FromSelect && modes_.hasCheck) {
    if (auto* m = findProp<GCheck>()) {
      if (m->isMutable()) *m = state;
    }
  }
  GBox* par = parent_ ? parent_->toBox() : nullptr;
  
  // on::action is :
  // - always triggered on a checkitem
  // - triggered on a radioitem only if it is checked
  GEventType etype = GEventType::Change;
  if (state || !par || !par->modes_.hasChoice) {
    etype = etype | GEventType::Action;
  }
  GTrigger cond(etype, (state ? 1 : 0));
  GEvent e(cond, GCore::elapsedTime());
  doFire(e);      // appellé sur les textitem qunad on les selecte !

  // if (state..) added
  if (state && from != FromChoice && par && par->modes_.hasChoice) {
    // unselect g so that choose() will actually do something
    modes_.checked = false;
    if (auto* m = par->choiceProp()) m->choose(*par, this);

    // select chidren of this box
    if (auto* b = toBox()) {
      for (auto& it : b->children()) {
        it->selectImpl(modes_.checked, FromChoice);
      }
    }
  }
  update();
}

// - - - Show and Update  - - - - - - - - - - - - - - - - - - - - - - - - - -

Gadget& Gadget::show(bool state) {
  if (modes_.shown == state) return *this;
  modes_.shown = state;
  // le faire meme si !state car ca affecte les parents
  updateLayout(Update::Needed);
  if ((eventmask_ & GEventType::Show) != 0) {
    GEvent e(state ? on::show : on::hide, GCore::elapsedTime());
    doFire(e);
  }
  return *this;
}

void Gadget::update() {
  if (!isShown()) return; 
  if (auto* top = getWindow()) top->update();
}

Gadget* Gadget::getUpdateParent() {
  Gadget* g = this;
  while (g) {
    if (g->isTopLevel() || g->layout_.activePos /* || g->modes_.hasFixedSize*/) return g;
    g = g->parent_;
  }
  return nullptr;
}

void Gadget::updateLayout(Update mode) {
  if (!isShown() && mode == Update::IfNeeded) return;

  auto* upd_root = getUpdateParent();
  if (!upd_root) return;

  if (mode == Update::Now) {      // sert pour control::paint()
    upd_root->computeSize(nullptr, true);
    upd_root->pack(nullptr);
    upd_root->update();
  }
  else {
    bool upd_css = (mode == Update::Css);
    auto* upd_win = upd_root->toWindow();  // is upd_root a window ?
    if (upd_win) {                         // needs to layout entire window
      GCore::updateLayout(upd_win, nullptr, upd_css);
    }
    else if ((upd_win = upd_root->getWindow())) {  // needs to layout part of the window
      GCore::updateLayout(upd_win, upd_root, upd_css);
    }
  }
}

// - - - Geometry - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

float Gadget::angleStart() const {
  return shape_ ? shape_->start : 0.f;
}

float Gadget::angleExtent() const {
  return shape_ ? shape_->extent : 0.f;
}

void Gadget::setAngleStart(float val) {
  if (!shape_) shape_ = new GadgetShape;
  shape_->start = val;
  update();
}

void Gadget::setAngleExtent(float val) {
  if (!shape_) shape_ = new GadgetShape;
  shape_->extent = val;
  update();
}

void Gadget::setShape(GBorderValue* b) {
  if (!shape_) shape_ = new GadgetShape;
  shape_->border = b;
}

GPoint Gadget::pos() const {
  if (parent_) return GPoint(frame_.x - parent_->frame_.x,
                             frame_.y - parent_->frame_.y);
  else return GPoint(frame_.x, frame_.y);
}

float Gadget::x() const {
  if (parent_) return frame_.x - parent_->frame_.x;     // if (parent_) ???
  else return frame_.x;
}

float Gadget::y() const {
  if (parent_) return frame_.y - parent_->frame_.y;
  else return frame_.y;
}

GPoint Gadget::screenPos() const {
  auto* w = getWindow();
  if (w) return GPoint(w->x() + frame_.x, w->y() + frame_.y);
  else return GPoint();
}

float Gadget::screenX() const {
  auto* w = getWindow();
  if (w) return w->x() + frame_.x;
  else return 0;
}

float Gadget::screenY() const {
  auto* w = getWindow();
  if (w) return w->y() + frame_.y;
  else return 0;
}

void Gadget::move(GPoint const& p) {
  prop<GPos>() = p;
}

void Gadget::move(float x, float y) {
  prop<GPos>().set(x,y);
}

void Gadget::moveImpl(float x, float y, bool callcb) {
  if (frame_.x == x && frame_.y == y) return;
  frame_.x = x;
  frame_.y = y;

  if (callcb && (eventmask_ & GEventType::Move) != 0) {
    GEvent e(GTrigger::move, GCore::elapsedTime());
    doFire(e);
  }
  if (layout_.passivePos) {
    if (auto* pos = findProp<GPos>()) pos->set(frame_.x, frame_.y);
  }
}

void Gadget::movedResizedCB(GRect const& initframe) {
  if ((eventmask_ & GEventType::Move) != 0
      && (frame_.x != initframe.x || frame_.y != initframe.y)) {
    GEvent e(GTrigger::move, GCore::elapsedTime());
    doFire(e);
  }

  if ((eventmask_ & GEventType::Resize) != 0
    && (frame_.w != initframe.w || frame_.h != initframe.h)) {
  GEvent e(GTrigger::resize, GCore::elapsedTime());
  doFire(e);
  }

  if (layout_.passivePos) {
    if (auto* pos = findProp<GPos>()) pos->set(frame_.x, frame_.y);
  }
}

void Gadget::resize(float w, float h) {   // !!NOTE : ne fixe plus flex a 0
  prop<GSize>().set(w, h, GSize::Pixels, GSize::Pixels);
}

void Gadget::resize(GDim const& d) {
  prop<GSize>().set(d.w, d.h, GSize::Pixels, GSize::Pixels);
}

int8_t Gadget::rowFlex(Gadget const* parent) {
  if (layout_.flex1 == layout_.flex2) return layout_.flex1;
  switch (layout_.flexType) {
    case 0:  // natural: gadget's w and h
      return layout_.flex1;
      break;
    case 1:  // self: gadgets' main and cross
      return isColumn() ? layout_.flex2 : layout_.flex1;
      break;
    default:  // parent: parent's  main and cross
      return (parent && parent->isColumn()) ? layout_.flex2 : layout_.flex1;
      break;
  };
}

int8_t Gadget::columnFlex(Gadget const* parent) {
  if (layout_.flex1 == layout_.flex2) return layout_.flex1;
  switch (layout_.flexType) {
    case 0:  // natural: gadget's w and h
      return layout_.flex2;
      break;
    case 1:  // self: gadgets' main and cross
      return isColumn() ? layout_.flex1 : layout_.flex2;
      break;
    default:  // parent: parent's  main and cross
      return (parent && parent->isColumn()) ? layout_.flex1 : layout_.flex2;
      break;
  };
}

void Gadget::setRowFlex(Gadget const* parent, int8_t val) {
  if (layout_.flex1 == layout_.flex2) {
    layout_.flex1 = val;
    return;
  }
  switch (layout_.flexType) {
    case 0:  // natural: gadget's w and h
      layout_.flex1 = val;
      break;
    case 1:  // self: gadgets' main and cross
      if (isColumn()) layout_.flex2 = val;
      else layout_.flex1 = val;
      break;
    default:  // parent: parent's  main and cross
      if (parent && parent->isColumn()) layout_.flex2 = val;
      else layout_.flex1 = val;
      break;
  };
}

void Gadget::setColumnFlex(Gadget const* parent, int8_t val) {
  if (layout_.flex1 == layout_.flex2) {
    layout_.flex1 = val;
    return;
  }
  switch (layout_.flexType) {
    case 0:  // natural: gadget's w and h
      layout_.flex2 = val;
      break;
    case 1:  // self: gadgets' main and cross
      if (isColumn()) layout_.flex1 = val;
      else layout_.flex2 = val;
      break;
    default:  // parent: parent's  main and cross
      if (parent && parent->isColumn()) layout_.flex1 = val;
      else layout_.flex2 = val;
      break;
  };
}

// computes the initial prefereed size (not the actual size!)
void Gadget::computeSize(GRender* parr, bool use_size) {
  if (!isShown()) return;

  GRender r(*this, parr);
  GDim objsize, size_init(frame_.w, frame_.h);

  // 2e iteration necessaire pour box contenant des multiline
  // !!! BUG: fait deconner les scrollbox: mauvais calcul, mettent scrrolbars a 0
  //int box_iter = 0;
  //BOX_AGAIN:
  if (GBox* box = toBox()) {
    //box_iter++;
    objsize = box->computeInsideSize(r);
  }

  if (r.margin_) {
    objsize.w += r.margin_->left + r.margin_->right;
    objsize.h += r.margin_->top + r.margin_->bottom;
  }
  if (r.border_) {
    objsize.w += r.border_->left + r.border_->right;
    objsize.h += r.border_->top  + r.border_->bottom;
  }

  // place pour les icones dans les menus, etc.
  if (r.icon_ && !toWindow()) {
    if (!parr || !parr->obj_->modes_.menuLayout) {
      objsize.w += r.font_.fontHeight();     // icon width will be fontHeight!
      if (r.text_ && !r.text_->empty()) objsize.w += r.gap_.mainspacing;
    }
  }

  if (r.text_) {
    // devrait pas faire la meme chose suivant que Size present ou pas !!!!!
    GDim ts;
    auto* attrlist = toTextItem() ? toTextItem()->attrlist_ : nullptr;
    if (modes_.multiLine) {
      // !!! PBM taille parent inconnue
      GBox* pere{};
      float hintw{};
      //if (parr && parr->obj_->frame_.w > 0) hintw = parr->obj_->frame_.w ;
      if (parr && (pere = r.obj_->getParent())) hintw = pere->childrenArea(*parr).w;
      if (hintw <= 0) hintw = r.size_.w;
      ts = GTextRender::layoutText(r.font_, modes_.richText, *r.text_,
                                   hintw, attrlist, nullptr);
      // cerr << "compsize: " <<  maxw
      // << " / " << (objsize.w+ts.w) << " " << (objsize.h+ts.h) << endl;
    }
    else {
      ts = r.font_.textSize(*r.text_, modes_.richText, attrlist);
    }
    objsize.w += ts.w;
    objsize.h += ts.h;  // must be added even if text is empty
  }
  // items must always have space for text
  else if (toItem()) {
    objsize.h += r.font_.fontHeight();  // must be added even if no text
  }

  if (r.hotkey_) {
    if (parr && parr->obj_->modes_.menuLayout) {
      objsize.w += r.font_.textSize(*r.hotkey_, false, nullptr).w;
    }
    else {
      objsize.w += r.font_.textSize(" ("+ *r.hotkey_ +")", false, nullptr).w;
    }
  }

  // the size isn't specified
  if (!use_size /* || r.flex_.mode == GFlex::NoSize*/) {
    frame_.w = objsize.w;
    frame_.h = objsize.h;
  }
  else {
    // flex < 0 means flex must not be changed
    if (layout_.flex1 >= 0) layout_.flex1 = r.flex_.flex1;
    if (layout_.flex2 >= 0) layout_.flex2 = r.flex_.flex2;
    layout_.flexType = r.flex_.mode;

    auto& sz = r.size_;

    switch (sz.wunit) {
      case GSize::Auto:     // auto-computed
        frame_.w = objsize.w;
        break;
      case GSize::Pixels:   // fixed size
        frame_.w = sz.w;
        break;
      case GSize::Chars: {   // fixed size in chars
        frame_.w = sz.w * r.font_.fontWidth();
      } break;
      case GSize::Percent:
        frame_.w = objsize.w;
        //layout_.wpercent = (uint8_t)sz.w;
        break;
        /*
         default: {   // fixed size in chars: use this char
         String s(1, sz.opt.wtype);
         frame.w = sz.w * font_->textSize(s, obj_.isFormatted()).w;
         } break; */
    }

    switch (sz.hunit) {
      case GSize::Auto:     // auto-computed
        frame_.h = objsize.h;
        break;
      case GSize::Pixels: // fixed size
        frame_.h = sz.h;
        break;
      case GSize::Chars:
        frame_.h = sz.h * r.font_.fontHeight();
        break;
      case GSize::Percent:
        frame_.h = objsize.h;
        //layout_.hpercent = int8_t(sz.h);
        break;
    }
  }

  // special case for windows: they cant be resized by their parent,
  // so flex means whether they adapt to their content
  if (auto* win = toWindow()) {
    if (r.flex_.flex1 > 0) frame_.w = objsize.w;
    if (r.flex_.flex2 > 0) frame_.h = objsize.h;
  }

  // can only grow if flex is < 0
  if (rowFlex(getParent()) < 0) {
    if (frame_.w < size_init.w) frame_.w = size_init.w;
  }
  if (columnFlex(getParent()) < 0) {
    if (frame_.h < size_init.h) frame_.h = size_init.h;
  }

  //if (box_iter == 1) goto BOX_AGAIN;  marche pas !!

  if ((eventmask_ & GEventType::Resize) != 0
      && (frame_.w != size_init.w || frame_.h != size_init.h)) {
    GEvent e(GTrigger::resize, GCore::elapsedTime());
    doFire(e);
  }
}

// - - - Pick & Pack - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

// supposes that ownPos is true
void Gadget::packPos(float parentx, float parenty) {
  // NB: Pos gadgets dont change the box size and may thus be clipped
  if (auto* pos = findProp<GPos>()) {
    auto initx = frame_.x, inity = frame_.y;
    frame_.x = parentx + pos->x();
    frame_.y = parenty + pos->y();
    if ((eventmask_ & GEventType::Move) != 0
        && (frame_.x != initx || frame_.y != inity)) {
      GEvent e(on::move, GCore::elapsedTime());
      doFire(e);
    }
  }
}

Gadget* Gadget::doPick(GPoint const& p, GPickMode& pm) {
  if (!modes_.shown || frame_.w <= 0.f || frame_.h <= 0.f) return nullptr;

  if (modes_.ignoreEvents && !pm.find_) {
    // callback?
    return nullptr;
  }

  // no reshape frame is shown
  if (layout_.reshapeFrame == 0 || pm.find_) {
    if (!frame_.contain(p)) return nullptr;

    auto* par = getParent();
    if (par && par->isPie() && !GGraphics::pickPie(frame_,angleStart(),angleExtent(),p)) {
      return nullptr;
    }
  }
  else {      // a reshape frame is shown
    if (isChecked()) {
      // check if inside the reshape frame
      if (!frame_.contain(p, GPrefs::resizeFrameShift()+GPrefs::resizeHandleSize()))
        return nullptr;
    }
    // not checked but the shape if very small and hard to reach
    // => add tolerance to ease selection
    else if (!frame_.contain(p, (frame_.w <= SmallShape || frame_.h <= SmallShape)
                             ? SmallShape : 0))
      return nullptr;
  }

  // the gadget has a shape
  if (shape_ && shape_->border && !shape_->border->pickBorder(*this, frame_, p)) {
    return nullptr;
  }

  // a predicate must be verified
  if (pm.pred_ && !(pm.pred_)(*this)) return nullptr;
  return this;
}


bool Gadget::pickHandle(GPoint const& pt, float x, float y) {
  auto handle = GPrefs::resizeHandleSize();
  return
  pt.x >= x-handle/2.f && pt.x <= x+handle/2.f
  && pt.y >= y-handle/2.f && pt.y <= y+handle/2.f;
}

GHandle Gadget::findHandle(GPoint const& wpos) {
  auto shift = GPrefs::resizeFrameShift();
  if (pickHandle(wpos, frame_.x -shift, frame_.y -shift))
    return GHandle::TopLeft;
  else if (pickHandle(wpos, frame_.x +frame_.w+shift, frame_.y -shift))
    return GHandle::TopRight;
  else if (pickHandle(wpos, frame_.x -shift, frame_.y +frame_.h +shift))
    return GHandle::BottomLeft;
  else if (pickHandle(wpos, frame_.x +frame_.w +shift, frame_.y +frame_.h +shift))
    return GHandle::BottomRight;
  else if (frame_.contain(wpos))
    return GHandle::Inside;
  else return GHandle::Outside;
}

// - - - Interactive Move & Resize - - - - - - - - - - - - - - - - - - - - - - -

void Gadget::reshapeInteractively() {
  auto* pos = findProp<GPos>();
  if (!pos || pos->type() == GPos::Type()) {
    *this << new GDragPos();
  }
  auto* size = findProp<GSize>();
  if (!size || size->type() == GSize::Type()) {
    *this << new GDragSize();
  }
}

void Gadget::reshapeInteractively(GMouseEvent& e) {
  reshapeInteractively();
  e.reshapeinter_ = true;
  e.redirectEvent(*this);
}

GHandle Gadget::handleDown(GPoint const& wpos, bool inter) {
  if (inter) return GHandle::BottomRight;
  else return findHandle(wpos);
}

void Gadget::handleUp(GPoint const&, GHandle const& loc) {
  updateLayout();
}

//NOTE: LineShape has its own version
void Gadget::handleDrag(float dx, float dy, GHandle const& loc) {
  GBox* par = getParent();
  if (!par) return;                       // !!!! pbm toplayers !!!
  //cerr << "reshapeDrag " << int(loc)<< " / " << dx << " " <<dy << endl;

  GPoint p;
  if (auto* pos = findProp<GPos>()) {
    p.set(pos->x() + dx, pos->y() + dy);
  }
  else {
    p.set(this->x() + dx, this->y() + dy);
  }
  
  switch (loc) {
    case GHandle::Inside:
      keepInside(*par, p);
      move(p);
    return;

    case GHandle::TopLeft: {
      keepTopLeftInParent(*par, p, dx, dy);
      float w = this->w() - dx, h = this->h() - dy;
      keepBottomRightInParent(*par, p, w, h);
      if (w > 0 && h > 0) {
        resize(w, h);
        move(p);
        updateLayout();
      }
    } return;

    case GHandle::TopRight: {
      keepTopLeftInParent(*par, p, dx, dy);
      float w = this->w() + dx, h = this->h() - dy;
      keepBottomRightInParent(*par, p, w, h);
      if (w > 0 && h > 0) {
        resize(w, h);
        move(x(), y()+dy);
        updateLayout();
      }
    } return;

    case GHandle::BottomLeft: {
      keepTopLeftInParent(*par, p, dx, dy);
      float w = this->w() - dx, h = this->h() + dy;
      keepBottomRightInParent(*par, p, w, h);
      if (w > 0 && h > 0) {
        resize(w, h);
        move(x()+dx, y());
        updateLayout();
      }
    } return;

    case GHandle::BottomRight: {
      float w = this->w() + dx, h = this->h() + dy;
      keepBottomRightInParent(*par, p, w, h);
      if (w > 0 && h > 0) {
        resize(w, h);
        updateLayout();
      }
    } return;

    default:
      return;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void Gadget::doPaint(GRenderGraphics& gr, GRender* parr) {
  GRender r(*this, parr);
  if (!r.paintGadget(gr, GRender::DrawAll)) return;

  if (isChecked() && layout_.reshapeFrame) paintHandleFrame(gr);
}

void Gadget::paintHandle(GRenderGraphics& g, float x, float y, float handle) {
  g.drawRect(x-handle/2, y-handle/2, handle, handle);
}

void Gadget::paintHandleFrame(GRenderGraphics& g) {
  auto shift = GPrefs::resizeFrameShift();
  auto handle = GPrefs::resizeHandleSize();

  GRect frame{
    frame_.x - shift, frame_.y - shift,
    frame_.w + 2*shift, frame_.h + 2*shift
  };

  GRect clip{
    frame_.x - shift-handle/2, frame_.y - shift-handle/2,
    frame_.w + 2*shift+handle, frame_.h + 2*shift+handle
  };
  g.setWinClip(clip);

  g.setColor(GPrefs::resizeFrameColor());
  g.unsetBgColor();
  g.drawRect(frame);
  g.setBgColor(GPrefs::resizeHandleColor());

  if (layout_.reshapeFrame & layout_.activeSize) {
    paintHandle(g, frame.x, frame.y, handle);
    paintHandle(g, frame.x + frame.w, frame.y, handle);
    paintHandle(g, frame.x, frame.y + frame.h, handle);
    paintHandle(g, frame.x + frame.w, frame.y + frame.h, handle);
  }
  g.setWinClip(frame_);
}

void Gadget::keepInside(GBox& par, GPoint& p) {
  auto handle = GPrefs::resizeHandleSize();
  if (p.x + w() < handle) {
    p.x = handle - w();
  }
  else if (p.x > par.w() - handle) {
    p.x = par.w() - handle;
  }
  if (p.y + h() < handle) {
    p.y = handle - h();
  }
  else if (p.y > par.h() - handle) {
    p.y = par.h() - handle;
  }
}

/*
void Gadget::keepInParent(GBox& par, GPoint& p, float w, float h) {
  auto handle = GPrefs::resizeHandleSize();
  if (p.x + w < handle) {
    p.x = handle - w;
  }
  else if (p.x > par.w() - handle) {
    p.x = par.w() - handle;
  }
  if (p.y + h < handle) {
    p.y = handle - h;
  }
  else if (p.y > par.h() - handle) {
    p.y = par.h() - handle;
  }
}*/

void Gadget::keepTopLeftInParent(GBox& par, GPoint& p, float& dx, float& dy) {
  if (p.x < 0) {
    p.x = 0;
    dx = p.x - this->x();
  }
  else if (p.x > par.w()) {
    p.x = par.w();
    dx = p.x - this->x();
  }
  if (p.y < 0) {
    p.y = 0;
    dy = p.y - this->y();
  }
  else if (p.y > par.h()) {
    p.y = par.h();
    dy = p.y - this->y();
  }
}

void Gadget::keepBottomRightInParent(GBox& par, GPoint& p, float& w, float& h) {
  if (p.x + w > par.w()) {
    w = par.w() - p.x;
  }
  if (p.y + h > par.h()) {
    h = par.h() - p.y;
  }
}

}
