//
//  Base class of GUI gadgets.
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Gadget_hpp
#define Guit_Gadget_hpp
#include <gobject.hpp>
#include <gflow.hpp>
#include <gproplist.hpp>
#include <gadgettype.hpp>
namespace guit {

/// Location of move/resize events when reshaping a Gadget.
enum struct GHandle : uint8_t {
  Outside, Inside,
  Left, Right, Top, Bottom,
  TopLeft, TopRight, BottomLeft, BottomRight,
  Point1, Point2  // for GLineShape
};


/// Constants for cloning Gadgets.
/// @see Gadget::clone().
enum struct GClone : uint8_t {
  Self = 0,          ///< Clones gadget.
  Children = 1<<0,   ///< Clones gadget and descendants.
  Props = 1<<1,      ///< Clones props of gadget (and descendants if Chidren is selected)
  ShareProps = 1<<2, ///< Shares props of gadget (and descendants if Chidren is selected)
  Sync = 1<<5        ///< Clones gadget and descendants, shares all props,  syncs views
};
GClone operator|(GClone,GClone);
uint8_t operator&(GClone,GClone);


/// Constants for reading/writing Gadgets.
/// @see Gadget::read() and Gadget::write().
enum struct GIOMode {
  Auto,
  Values,
  Content
};


/// Base class of Graphical Objects.
/// See Gadget documentation page: @ref gadgetdoc
class Gadget : public GObject {
protected:

  /// Creates a new gadget.
  /// _css_and_title_ specifies CCS styles and the title/value of the gadget (see below).
  /// _css_and_title_ is in the form:
  /// @code
  ///    ".class1, .class2, [...,] #id {inline_style} title_or_value"
  /// @endcode
  /// where all fields are optional and:
  /// - class1, class2, etc. are CSS styles shared by multiple gadgets
  /// - id is a CSS style that is specific to this gadget
  /// - { inline_style } is an inline CSS style, e.g. { color: red; border: squared }
  /// - title_or_value is either the initial title or value of the gadget (the second case applies to gadgets
  ///  having a value, see hasValue()). For instance, title_or_value can be:
  ///    # The title of a Box, Window, Dialog, Label, Button, Item, etc.
  ///    # The value of a TextField, TextArea, Slider, ChartItem, etc.
  Gadget(GadgetType& css_and_title);

  Gadget(GadgetType&, GString const& css_and_title);

public:
  using IsGadget = bool;
  using IsGadgetOrVarProp = bool;
  using IsGadgetChild = bool;

  /// Destructor: don't use the delete operator!.
  /// Objects are auto-deleted, thus delete operator should never be called (it just issues a warning).
  /// To delete an object of class MyClass pointed by p, just do:
  /// @code
  ///  gptr<GObject> p2 = p;
  ///  p2 = nullptr;
  /// @endcode
  /// @see Gadget and gptr.
  virtual ~Gadget();

  /// Clones the gadget according to GClone argument.
  virtual Gadget& clone(GClone const&) = 0;

  /// Returns gadget's meta-class.
  /// Both functions return the gadget's type, but Type() is a static method and type() a virtual method.
  /// These functions are redefined in instantiable subclasses.
  /// @{
  using   GType = GadgetType;
  static  GType& Type();
  virtual GType& type() const {return Type();}
  /// @}

  /// Returns the name and ID of the gadget's type.
  /// @{
  GString typeName() const override {return type().name_;}
  GId typeId() const {return type().id_;}
  /// @}

  /// Class conversions.
  /// @{
  Gadget*              toGadget() override {return this;}
  Gadget const*        toGadget() const override {return this;}
  virtual GItem*       toItem()       {return nullptr;}
  virtual GTextItem*   toTextItem()   {return nullptr;}
  virtual GValueItem*  toValueItem()  {return nullptr;}
  virtual GShape*      toShape()      {return nullptr;}
  virtual GChoiceBox*  toChoiceBox()  {return nullptr;}
  virtual GScrollBox*  toScrollBox()  {return nullptr;}
  virtual GChartBox*   toChartBox()   {return nullptr;}
  virtual GWindow*     toWindow()     {return nullptr;}
  virtual GMenu*       toMenu()       {return nullptr;}
  virtual GMenuBar*    toMenuBar()    {return nullptr;}
  virtual GTable*      toTable()      {return nullptr;}
  virtual GTRow*       toTRow()       {return nullptr;}
  virtual GTHead*      toTHead()      {return nullptr;}
  /// @}

  /// Blinks the gadget, optionally showing a message.
  /// Shows a message if _message_ is not empty. The duration is in ms.
  void blink(GString const& message = "", GTime duration = 350);

  /// Returns true if the gadget contains this prop.
  /// @{
  virtual bool contain(GProp* prop) const;
  virtual bool contain(GProp& prop) const {return contain(&prop);}
  /// @}

  /// Searches if the gadget contains a prop of this class or a derived class.
  /// The first version returns the prop as a GProp* if it is found, and null otherwise.
  /// In addition, the second version casts the returned prop to the desired type.
  ///
  /// @note The returned prop may be immutable (ie if GVarProp::isMutable() return false).
  /// For instance, predefined variables such as GColor::red or GFont::bold are immutable.
  ///
  /// @note Props in inline styles (e.g. Button( "{color:red; font:bold} DoIt" ) belong to the gadget,
  /// they are added to it in the same way as if using operator<<().
  /// @{
  virtual GProp* findProp(GPropType&) const;

  template <class T>
  T* findProp() const {return dynamic_cast<T*>(findProp(T::Type()));}
  /// @}

  /// Searches a prop of this class in the gadget and the CSS, and creates it if needed.
  /// This method always returns a mutable prop of the desired type:
  /// - first it searches if the gadget contains this prop by calling find()
  /// - then, if 'in_css' is true, it also searches in CSSs (not the default because somewhat CPU intensive)
  /// - finally, if no corresponding prop is found, it created it and adds it to the gadget
  ///
  /// @note If the found prop is immutable (see find()), it is duplicated and the copy is added to the
  /// gadget then returned
  /// @see findProp() and cssProps()
  /// @{
  virtual GProp& prop(GPropType&, bool in_css = false);

  template <class T>
  T& prop(bool in_css = false) {return dynamic_cast<T&>(prop(T::Type(), in_css));}
  /// @}

  /// Returns all props.
  /// @note Does not return the props relative to the gadget that are defined in CSS styles
  /// (see cssProps() for this purpose), except those that are defined in inline styles, eg in:
  /// Button( "{color:red; font:bold} DoIt" )
  GPropList const& props() const {return props_;}

  /// Returns the props relative to this gadget in CSSs.
  /// @note Does not return the props that belong to the gadget (see props() for this purpose),
  /// except those that are defined in inline styles, eg in: Button( "{color:red; font:bold} DoIt" )
  void cssProps(class GPropMap&);

  /// Returns the text of the gadget (if any).
  /// Returns the text of gadgets that display text, e.g. Item, Label, Button, TextField, etc.,
  /// and an empty string otherwise.
  virtual GString text() const;

  /// Adds/replaces props.
  /// operator<<() provides a more convenient API:  a << b << c  adds b and c to a.
  /// The appropriate adding function is automatically called.
  ///
  /// @note These methods do not just adds props, they also *replaces* them if the added prop
  /// has the same role (GProp::role()) as an existing gadget's prop. For instance, if the gadget has
  /// a GColor prop, adding a new GColor to it will replace the previous one.
  ///
  /// @note Replacing a prop will delete the old prop if it does not belong to another gadget and is not
  /// gptr-pointed elsewhere (see gptr<> smart pointer)
  ///
  /// @note When a GPropList (list of props) is given as an argument, each of it elements is added
  /// in sequence.
  /// @see Gadget, GProp, GCond (conditional props).
  /// @{
  virtual void add(GString const& str);
  void add(GProp* prop);
  void add(GProp& prop) {addProp(prop,true);}
  void add(GPropList const& proplist);
  /// @}

  /// Adds a callback function or a state transitions.
  /// operator<<() provides a more convenient API:  a << b << c  adds b and c to a.
  /// The appropriate adding function is automatically called.
  ///
  /// @note The default on::action triggers of the gadget is used. For instance, the callback will be called
  /// if a GButton is clicked, if a GCheckItem is checked or unchecked, if a GRadioItem belonging to
  /// a box that performs exclusive selection is checked.
  /// @{
  void add(class GStateTransition&);
  void addFun(GFunction const&);
  void addFun(GEventFunction const&);
  void addFun(GCaller&);
  /// @}

  /// Adds a menu, a window, a dialog box to the gadget.
  /// operator<<() provides a more convenient API:  a << b << c  adds b and c to a.
  /// The appropriate adding function is automatically called.
  ///
  /// If the gadget is a GItem or GButton, the GMenu, GWindow, GDialog given as an argument
  /// will be automatically opened when pressing the mouse on the item (GMenu) or clicking it
  /// (GWindow, GDialog).
  ///
  /// @note Only one GMenu can be added to a gadget. If another menu is added, it will replace
  /// the previous one.
  /// @{
  virtual void add(GMenu* child);
  virtual void add(GMenu& child) {add(&child);}
  virtual void add(GWindow*);
  virtual void add(GWindow& child) {add(&child);}
  /// @}

  /// Removes a prop (including conditional props and callbacks).
  /// Returns true if the prop was actually removed (i.e. if it was belonging to the gadget).
  ///
  /// @note Removing a prop will delete it if it does not belong to another gadget and is not
  /// gptr-pointed elsewhere (see gptr<> smart pointer)
  /// @see Gadget, GProp, GCond (conditional props).
  /// @{
  bool remove(GProp* prop);
  bool remove(GProp& prop);
  virtual bool remove(GTrigger&);
  virtual bool remove(GPropType&);
  /// @}

  /// Removes all props (including conditional props and callbacks).
  /// @note Beware that the removed props will be auto-deleted if not gptr-referenced elsewhere (see Gadget).
  virtual void removeProps();

  /// Removes all props and children.
  /// This method also removes children if the gadget has children (i.e. if the class derives from GBox).
  virtual void cleanup();


  /// Returns true if gadget displays multiline text.
  bool isMultiLine() const {return modes_.multiLine;}

  /// Checks/changes whether HTML-like tags are interpreted when displaying text.
  /// HTML-like tags such as \<b>, \<i>, etc. will be interpreted when rendering text.
  /// This option is true by default, except for gadgets that edit text (e.g. TextField and TextArea).
  /// @{
  bool isRichText() const {return modes_.richText;}
  Gadget& setRichText(bool = true);
  /// @}

  /// Checks/changes whether the gadget catches events.
  /// The gadget will receive events that should have been received by its descendants.
  /// The received GEvent will be set so that:
  /// - GEvent::gadget()  will return the gadget
  /// - GEvent::item() will return the descendant that should have received the event.
  /// @{
  Gadget& catchEvents(bool = true);
  bool isCatchingEvents() const {return modes_.catchEvents;}
  /// @}

  /// Checks/changes whether the gadget ignores events.
  /// The gadget and its descendants will then ignore all events.
  /// @{
  Gadget& ignoreEvents(bool = true);
  bool isIgnoringEvents() const {return modes_.ignoreEvents;}
  /// @}

  /// Checks/changes whether the gadget is checked (ie selected).
  /// check() has not effect if isCheckable() is false; see setCheckable().
  /// @{
  virtual Gadget& check(bool = true);
  bool isChecked() const {return modes_.checked;}
  /// @}

  /// Checks/changes whether the gadget is checkable (ie can be selected).
  /// setCheckable(true) makes any gadget checkable. Some gadgets (e.g. GCheckItem, GRadioItem)
  /// are checkable by default, but most others are not.
  /// @{
  Gadget& setCheckable(bool = true);
  bool isCheckable() const {return modes_.checkable;}
  /// @}

  /// Checks/change whether the gadget is enabled.
  /// - enable() sets isSelfEnabled() to true or false
  /// - isTrulyEnabled() is true if the gadget and all its ancestors ar isSelfEnabled().
  /// @{
  virtual Gadget& enable(bool = true);
  bool isSelfEnabled() const {return modes_.enabled;}
  bool isTrulyEnabled() const {return settings_.trulyEnabled;}
  /// @}

  /// Checks or change whether the gadget is shown.
  /// isShown() does mean that the gadget is shown, but that it will be shown if all its ancesters
  /// are shown
  /// @{
  virtual Gadget& show(bool = true);
  bool isShown() const {return modes_.shown;}
  /// @}

  /// Notifies the toolkit that this gadget needs to be repainted.
  /// The gadget is not repainted immediately, but after executing callback functions.
  /// The doPaint() method is then called.
  virtual void update();

  /// Layout update modes (see updateLayout()).
  enum struct Update {IfNeeded, Needed, Now, Css};

  /// Notifies the toolkit that the layout needs to be updated.
  /// By default (UpdateIfNeeded mode) the layout is not updated immediately, but after executing
  /// callback functions, and before repainting the gadget by calling doPaint().
  ///
  /// Possible arguments are:
  /// - IfNeeded (default): the layout is updated if the toolkit thinks this is needed
  /// - Needed: the layout is always updated
  /// - Now: the layout is updated immediately. Sometimes needed but CPU inefficient
  virtual void updateLayout(Update = Update::IfNeeded);

  /// Returns the parent gadget.
  /// - getParent() returns the gadget's parent if it derives from GBox, and null if no parent or not a box.
  /// - getPseudoParent() returns the gadget's parent whatever its type.
  /// Displayed gadgets have a non-null parent that derives from GBox, except toplevel gadgets
  /// (GMenu, GWindow, GDialog, etc.) which parent can be null or can be any gadget.
  /// @{
  GBox* getParent() const {return parent_ ? parent_->toBox() : nullptr;}
  virtual Gadget* getPseudoParent() const {return parent_;}
  ///@}

  /// Returns the GWindow that contains the gadget (if any, null otherwise).
  virtual GWindow* getWindow() const;

  /// Returns true if the gadget is a Window or a Menu.
  virtual bool isTopLevel() const {return false;}

  /// Returns true if the gadget has a value.
  /// Gadgets such as TextField, Slider, ChartItem, ChoiceBox, OptionBox, ComboBox have a value.
  /// This methods then returns true, and false otherwise.
  /// @see stringValue() and setValue().
  virtual bool hasValue() const {return false;}

  /// Returns the prop that stores the gadget's value.
  /// If the gadget has a value (see hasValue()) this method returns the prop storing the value.
  /// Otherwise it returns null.
  virtual GProp* valueProp() {return nullptr;}

  /// Returns the gadget's value as a string.
  /// If the gadget has a value (see hasValue()) this method returns the gadget value converted
  /// to a string. Otherwise it returns the empty string.
  virtual GString stringValue() const {return emptystring_;}

  /// Changes the gadget's value.
  /// If the gadget has a value (see hasValue()) this method changes the gadget value after
  /// converting _value_ to the appropriate type. Otherwise it has no effect.
  virtual void setValue(GString const& value) {}

  /// Clears the values of the gadget and its descendants.
  /// Clearing means setting to the default value (0 for numbers, the empty string for strings, etc.)
  virtual void clearValues();

  /// Reads the values or the children of this gadget recursively.
  /// @{
  bool read(GString const& filename, GIOMode = GIOMode::Auto);
  bool read(std::istream&, GIOMode = GIOMode::Auto);
  //friend std::istream& operator>>(std::istream&, Gadget&);
  /// @}

  /// Writes the values or the children  of this gadget recursively.
  /// @{
  bool write(GString const& filename, GIOMode) const;
  bool write(std::ostream&, GIOMode) const;
  //friend std::ostream& operator<<(std::ostream&, Gadget const&);
  /// @}

  /// Returns the direction of the gadget.
  /// The direction can be changed by adding a GFlow prop to the gadget.
  /// @{
  bool isColumn() const {return modes_.dir == GFlow::Column;}
  bool isRow() const {return modes_.dir == GFlow::Row;}
  bool isPie() const {return modes_.dir == GFlow::Pie;}
  /// @}

  /// Changes the position (in pixels) of the gadget.
  /// This function is a shortcut that changes the value of the gadget's GPos prop (and adds a GPos
  /// to the gadget if it doesn't already have one).  Importantly, the gadget won't be auto laid-out
  /// any longer if this function is called (or a GPos is directly added to the gadget).
  ///
  /// Gadgets that are automatically laid out don't need a GPos. Gadgets that are positioned
  /// manually need a GPos that specifies their position and (optionally) their zlayer().
  ///
  /// The position is relative to the parent gadget, or to the screen if the gadget is a GWindow.
  /// The zlayer() of auto-laid out gadgets is 0, the zlayer() of other gadgets is > 0 (see GPos).
  ///
  /// A GPos prop can be directly added to a gadget as follows:
  /// @code
  ///    auto& pos = Pos();
  ///    auto& win = Window("My Window") << pos;
  ///    pos.set(100,100);
  ///    ...
  ///    pos.set(200,200);
  /// @endcode
  /// Adding a GPos rather than calling move() allows:
  /// - controlling the zlayer of the gadget
  /// - better performance if the gadget is frequently moved
  /// - sharing the GPos with other gadgets (which are then synced)
  /// @{
  virtual void move(GPoint const&);
  virtual void move(float x, float y);
  /// @}

  /// Changes the preferred size (in pixels) of the gadget.
  /// This function is a shortcut that changes the value of the gadgte's GSIze prop (and adds a GSize
  /// to the gadget if it doesn't already have one). The actual size of the gadget also depends
  /// on its GFlex prop and the GAlign prop of its parent.
  ///
  /// The GSize prop can also be added directly by using the << operator (see move() for an example).
  /// Adding a GSize rather than calling resize() allows:
  /// - specifying the unit of w and h, or if they are computed automatically
  /// - better performance if the gadget is frequently resized
  /// - sharing the GSize with other gadgets (which are then synced)
  /// @{
  virtual void resize(float w, float h);
  virtual void resize(GDim const&);
  /// @}

  /// Allows reshaping a gadget interactivelly.
  /// Adds the appropriate GPos and GSize props to the gadget so that it will display handles
  /// allowing to reshape it when it is checked. "Reshaping" both means changing the position
  /// and the size (e.g. when dragging left-side handles).
  /// The second version of this function allows reshaping a gadget just after instantiating it. The
  /// GMouseEvent argument typically corresponds to a mouse down event in a gadget that serves
  /// as a canvas for drawing other gadgets (or geometrical shapes, see GShape).
  /// @{
  virtual void reshapeInteractively();
  virtual void reshapeInteractively(GMouseEvent&);
  /// @}

  /// Returns the position of the gadget in its parent.
  /// Return the position on the screen if the gadget is a GWindow,
  /// @{
  virtual GPoint pos() const;
  virtual float x() const;
  virtual float y() const;
  /// @}

  /// Returns the z-layer of the gadget.
  /// The z-layer is 0 if the gadget is auto laid-out and > 0 otherwise (see GPos prop).
  int zlayer() const {return layout_.layer;}

  /// Returns the position of the gadget in its containing window.
  /// Return 0 if the gadget is a GWindow.
  /// @{
  GPoint wpos() const {return GPoint(frame_.x,frame_.y);}
  float wx() const {return frame_.x;}
  float wy() const {return frame_.y;}
  /// @}

  /// Returns the position of the gadget on the screen.
  /// @{
  GPoint screenPos() const;
  float screenX() const;
  float screenY() const;
  /// @}

  /// Returns the dimension (size) of the gadget.
  /// @{
  float w() const {return frame_.w;}
  float h() const {return frame_.h;}
  GDim dim() const {return GDim(frame_.w, frame_.h);}
  /// @}

  /// Returns the frame (position and size) of the gadget in window coordinates.
  GRect const& wframe() {return frame_;}

  /// Returns/changes the starting angle and the extent when painting gadgets as pie slices.
  /// GItem gadgets are drawn as pie slices when contained in a box with a GFlow::pie prop.
  /// This is typically used in pie charts (see GChart) and pie menus (see GPieMenu).
  /// @{
  float angleStart() const;
  float angleExtent() const;
  void  setAngleStart(float);
  void  setAngleExtent(float);
  /// @}

  /// Placement of the gadget in a GBox parent.
  /// When adedd to a GBox, the placement of the gadget depends on its place() method:
  /// - Place::Content means it is added to the content area of the box
  /// - other values means it is added to the border of the box (e.g. Place::Top for a title or a menubar)
  /// @{
  enum struct Place : int8_t {
    Content = 0,
    Top = -10, Bottom = -20, Left = -30, Right = -40, ToolBar = -50, ScrollBar = -60
  };
  virtual Place place() const {return Place::Content;}
  ///@}

  bool unclippedLayers() const {return modes_.unclippedLayers;}
  Gadget& setUnclippedLayers(bool = true);
  
  /// Changes point to keep the gadget inside an ancestor when moving or resizing it.
  virtual void keepInside(GBox& box, GPoint& point);

  /// Returns the handle that contains this position in window corrdinates.
  virtual GHandle findHandle(GPoint const& wpos);

  /// Returns true if the gadget is currently active (eg pressed).
  bool isActive() const {return modes_.active;}

  /// Ckecks/changes whether the gadget can be active.
  /// @{
  bool isActivable() const {return modes_.activable;}
  Gadget& setActivable(bool = true);
  /// @}

  /// Ckecks/changes whether the gadget is active on mouse down rather than mouse up.
  /// @{
  bool isDownActivable() const {return modes_.downActivable;}
  Gadget& setDownActivable(bool = true);
  /// @}

  /// Returns true if the gadget has/can have the focus.
  /// @{
  bool isFocus() const {return modes_.focus;}
  bool isFocusable() const {return modes_.focusable;}
  /// @}

  /// Ckecks/changes whether the gadget is an internal (implementation-specific) object.
  /// internal objects are ignored when inspecting, reading or writting gadgets.
  /// @{
  bool isInternal() const {return modes_.internal;}
  void setInternal() {modes_.internal = 1;}
  /// @}

  /// Returns all gadget modes.
  GadgetModes const& modes() const {return modes_;}

  /// Returns the events (ORed mask) that the gadget takes into account.
  GEventType const& eventMask() const {return eventmask_;}

  /// Fires callbacks and conditional props (see GCond) that match this event.
  virtual void doFire(GEvent&);

  /// Paints the gadget.
  /// Repaints the gadget then fires on::paint callbacks. For instance, mycallback can be called
  /// to draw on the gadget as follows;
  /// @code
  ///     mygadget << on::paint / [this](GEvent& e){mycallback(e);};
  /// @endcode
  /// @see Graphics.
  virtual void doPaint(GRenderGraphics&, GRender* parent_render);

  /// Returns the gadget the gadget that contains this point.
  /// The point is in window coordinates.
  /// @see GBox::childAtPos() and GBox::childAtWPos()
  virtual Gadget* doPick(GPoint const& wpos, class GPickMode&);

  /// Called when the gadget is added to a parent.
  virtual void onAdd(GBox* parent) {}

  /// Called when the gadget is removed from a parent.
  virtual void onRemove(GBox* parent) {}

  /// @internal
  /// Guit provides access to all fields, but beware using them is at your own risk.
  /// @{
  class GadgetStyle* style() const {return style_;}
  virtual void init(GRender* parrender);
  virtual void initClone(Gadget& from, GClone const&);
  virtual void addProp(GProp&, bool add_to_list);
  virtual void addProp2(GProp&, bool add_to_list);
  virtual bool removeProp(GProp&, bool rem_from_proplist);
  virtual bool removeProp2(GProp&, bool rem_from_proplist);
  void removeNotifiers(void*) override;
  virtual void addOpener(Gadget& opened, GTrigger&, int btn, bool autoplace);

  virtual void readValues(std::istream&, class GIOBuffer&);
  virtual void writeValues(std::ostream&, class GIOBuffer&) const;

  virtual void readAll(std::istream&, class GIOBuffer&);
  virtual void writeAll(std::ostream&, class GIOBuffer&) const;

  virtual void readAux(std::istream&, class GIOBuffer&) {}
  virtual void writeAux(std::ostream&, class GIOBuffer&) const {}

  int8_t rowFlex(Gadget const* parent);
  int8_t columnFlex(Gadget const* parent);
  void setRowFlex(Gadget const* parent, int8_t val);
  void setColumnFlex(Gadget const* parent, int8_t val);

  enum SelectFrom {FromButton, FromChoice, FromSelect};
  // internal fct for (un)selecting gadget.
  virtual void selectImpl(bool state, SelectFrom from);

  // finds the ancestor that must be used for updating this gadget.
  virtual Gadget* getUpdateParent();

  // called by update() to compute preferred size.
  virtual void computeSize(GRender* parent_render, bool use_size_prop);

  // called by update() to lay out gadget and children.
  virtual void pack(GRender* parent_render) {}

  // called by update() to lay out gadgets that have a GPos
  virtual void packPos(float parentx, float parenty);

  // called by move() or GPos prop to move gadget.
  virtual void moveImpl(float wx, float wy, bool callcb);

  // calls on::move and/or or::resize callbacks.
  virtual void movedResizedCB(GRect const& initialframe);

  void setShape(class GBorderValue*);

  // called when using handles to resize the gadget interactively.
  // note: also changes the position when dragging left resize handles
  virtual GHandle handleDown(GPoint const& wpos, bool inter);
  virtual void handleDrag(float dx, float dy, GHandle const&);
  virtual void handleUp(GPoint const& wpos, GHandle const&);

  // paints handle frame when resizing the gadget interactively.
  virtual void paintHandleFrame(GRenderGraphics&);
  void paintHandle(GRenderGraphics& g, float x, float y, float handle);
  bool pickHandle(GPoint const& pt, float x, float y);

  // keep the gadget inside its parent when resizing it.
  virtual void keepTopLeftInParent(GBox& par, GPoint&, float& dx, float& dy);
  virtual void keepBottomRightInParent(GBox& par, GPoint& p, float& w, float& h);

  static const GString emptystring_;
  // order optimized for memory occupation
  GadgetSettings settings_{};
  GadgetModes modes_{};
  GadgetLayout layout_{};
  GEventType eventmask_{};
  GRect frame_;
  Gadget* parent_{};
  GPropList props_;
  gptr<class GadgetStyle> style_;
  GadgetShape* shape_{};
  /// @}
};


/// List of Gadgets.
using GadgetList = std::vector<gptr<Gadget>>;

}
#endif


#include <gadd.hpp>
