//
// Shape gadgets
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <sstream>
#include <giobuffer.hpp>
#include <gevent.hpp>
#include <gshape.hpp>
#include <gtrigger.hpp>
#include <gfloat.hpp>
#include <gbox.hpp>
#include <ggraphics.hpp>
#include <gcontrol.hpp>
#include <gstyle.hpp>
#include <grender.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_GADGET(RectShape)
GIMPLEMENT_GADGET(OvalShape)
GIMPLEMENT_GADGET(ArcShape)
GIMPLEMENT_GADGET(LineShape)
GIMPLEMENT_GADGET(FreehandShape)

GShape::GType& GShape::Type() {
  static auto& t = *new GShape::GType("Shape");
  return t;
}

GShape::~GShape() = default;


GShape::GShape(GadgetType& t, GString const& spec) : Gadget(t, spec) {
  eventmask_ = eventmask_ | GEventType::MouseButton;
}

GShape::GShape(GString const& spec) : GShape(Type(), spec) {}

Gadget& GShape::clone(GClone const& c) {
  auto* g = new GShape(Type(), "");
  g->initClone(*this, c);
  return *g;
}

GPos& GShape::posProp() {
  // Gadget::prop() recupere la valeur du style si definie
  if (!pos_) return prop<GPos>();
  else return *pos_;
}

GSize& GShape::sizeProp() {
  if (!size_) return prop<GSize>();
  else return *size_;
}

GProp* GShape::findProp(GPropType& t) const {
  if (t.role() == GPropType::Role::Pos)
    return pos_;
  else if (t.role() == GPropType::Role::Size)
    return size_;
  else
    return Gadget::findProp(t);
}

GProp& GShape::prop(GPropType& t, bool compute_style) {
  if (t.role() == GPropType::Role::Pos) {
    // doesnt work because prop<>() is not virtual!
    // would in fact call GShape::prop
    // return Gadget::prop<GPos>();
    return Gadget::prop(GPos::Type(), compute_style);
  }
  //else if (t.role() == GSize::Class().role()) {
  else if (t.role() == GPropType::Role::Size) {
    return Gadget::prop(GSize::Type(), compute_style);
  }
  else return Gadget::prop(t, compute_style);
}

void GShape::addProp(GProp& prop, bool add_to_list) {
  if (auto* pp = prop.toPos()) {
    pos_ = pp;
    addProp2(*pp, add_to_list);
    pp->startNotifyGadget(*this);
  }
  else if (auto* ps = prop.toSize()) {
    size_ = ps;
    addProp2(*ps, add_to_list);
    ps->startNotifyGadget(*this);
  }
  else Gadget::addProp(prop, add_to_list);
}

void GShape::onAdd(GBox* parent) {
  //            en fait ca devrait etre fait pour tous les gadgets !!!
  if (pos_) pos_->updateGadget(*this);
}

void GShape::move(float x, float y) {
  if (!pos_) add(new GPos());
  pos_->set(x,y);
}

void GShape::resize(float w, float h) {
  if (!size_) add(new GSize());
  size_->set(w, h, GSize::Pixels, GSize::Pixels);
}

void GShape::resize(GDim const& s) {
  if (!size_) add(new GSize());
  size_->set(s.w, s.h, GSize::Pixels, GSize::Pixels);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GRectShape::GRectShape(GadgetType& t, GString const& spec)
: GShape(t, spec) {}

GRectShape::GRectShape(GString const& spec)
: GShape(Type(), spec) {}

GRectShape& GRectShape::clone(GClone const& c) {
  auto* g = new GRectShape;
  g->initClone(*this, c);   // link pos_ and size_ !!!
  return *g;
}

void GRectShape::doPaint(GRenderGraphics& g, GRender* parr) {
  GRender r(*this, parr);
  if (!r.paintGadget(g, GRender::DrawNone)) return;   // DrawMargin
  g.setColor(r.color_);
  g.setBgColor(r.background_);
  g.setThickness(1); //r.border_->thickness);
  g.drawRect(frame_);
  if (isChecked() && layout_.reshapeFrame) paintHandleFrame(g);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GOvalShape::GOvalShape(GadgetType& t, GString const& spec)
: GRectShape(t, spec) {}

GOvalShape::GOvalShape(GString const& spec)
: GRectShape(Type(), spec) {}

GOvalShape& GOvalShape::clone(GClone const& c) {
  auto* g = new GOvalShape;
  g->initClone(*this, c);
  return *g;
}

void GOvalShape::doPaint(GRenderGraphics& g, GRender* parr) {
  GRender r(*this, parr);
  if (!r.paintGadget(g, GRender::DrawNone)) return;   // DrawMargin
  g.setColor(r.color_);
  g.setBgColor(r.background_);
  g.setThickness(1); //r.border_->thickness);
  g.drawOval(frame_);
  if (isChecked() && layout_.reshapeFrame) paintHandleFrame(g);
}

Gadget* GOvalShape::doPick(GPoint const& pt, GPickMode& pm) {
  auto* found = GShape::doPick(pt, pm);
  if (!found) return nullptr;
  else if (isChecked() && layout_.reshapeFrame) return this;
  else if (GGraphics::pickOval(frame_, pt)) return this;
  else return nullptr;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GArcShape::GArcShape(GadgetType& t, GString const& spec) :
GOvalShape(t, spec),
start_(new GFloat(0)),
extent_(new GFloat(360)) {
}

GArcShape::GArcShape(GString const& spec) :
GArcShape(Type(), spec) {}

GArcShape& GArcShape::clone(GClone const& cm) {
  auto* g = new GArcShape;
  // link start extents pos size   !!!
  g->initClone(*this, cm);
  return *g;
}

float GArcShape::start()  const {return start_->value();}
float GArcShape::extent() const {return extent_->value();}

GArcShape& GArcShape::setStart(float start) {
  start_->set(start);
  update();
  return *this;
}

GArcShape& GArcShape::setExtent(float extent) {
  extent_->set(extent);
  update();
  return *this;
}

void GArcShape::doPaint(GRenderGraphics& g, GRender* parr) {
  GRender r(*this, parr);
  if (!r.paintGadget(g, GRender::DrawNone)) return;   // DrawMargin
  g.setColor(r.color_);
  g.setBgColor(r.background_);
  g.setThickness(1); //r.border_->thickness);
  g.drawPie(frame_, start_->value(), extent_->value());   // !!!!
  if (isChecked() && layout_.reshapeFrame) paintHandleFrame(g);
}

Gadget* GArcShape::doPick(GPoint const& pt, GPickMode& pm) {
  auto* found = GShape::doPick(pt, pm);
  if (!found) return nullptr;
  else if (isChecked() && layout_.reshapeFrame) return this;
  else if (GGraphics::pickPie(frame_, *start_*360, *extent_*360, pt)) return this;
  else return nullptr;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GLineShape::GLineShape(GadgetType& t, GString const& spec)
: GShape(t, spec) {
  changePosProp(*new GPos());
  changePos2Prop(*new GPos());
}

GLineShape::GLineShape(GString const& spec)
: GLineShape(Type(), spec) {}

GLineShape& GLineShape::clone(GClone const& cm) {
  auto* g = new GLineShape;
  g->initClone(*this, cm);
  g->pos2_ = pos2_;
  return *g;
}

void GLineShape::onAdd(GBox* parent) {
  updateFrame();
}

void GLineShape::reshapeInteractively() {
  if (pos_->type() == GPos::Type())   // role() ?
    changePosProp(*new GDragPos());
  if (pos2_->type() == GPos::Type())   // role() ?
    changePos2Prop(*new GDragPos());
}

void GLineShape::reshapeInteractively(GMouseEvent& e) {
  reshapeInteractively();
  e.reshapeinter_ = true;
  e.redirectEvent(*this);
}

GPos& GLineShape::pos2Prop() {return *pos2_;}

GLineShape& GLineShape::changePosProp(GPos& p) {
  if (&p == pos_) return *this;
  pos_ = p;
  p << [this]{updateFrame();};
  addProp2(p, true);
  updateFrame();
  p.startNotifyGadget(*this);
  return *this;
}

GLineShape& GLineShape::changePos2Prop(GPos& p) {
  if (&p == pos2_) return *this;
  pos2_ = p;
  p << [this]{updateFrame();};
  updateFrame();
  p.startNotifyGadget(*this);
  return *this;
}

void GLineShape::addProp(GProp& prop, bool add_to_list) {
  if (auto* pp = prop.toPos()) {
    changePosProp(*pp);
  }
  else if (auto* ps = prop.toSize()) {
    size_ = ps;
    addProp2(*ps, add_to_list);
    updateFrame();
    ps->startNotifyGadget(*this);
  }
  else Gadget::addProp(prop, add_to_list);
}

/*
void GLineShape::resize(float w, float h) {
  if (!size_) addProp(*new GSize(), true);
  size_->set(w, h, GSize::Pixels, GSize::Pixels);
}

void GLineShape::resize(GDim const& s) {
  if (!size_) addProp(*new GSize(), true);
  size_->set(s.w, s.h, GSize::Pixels, GSize::Pixels);
}
*/

void GLineShape::move(float x, float y) {
  auto dx = x - pos_->x();
  auto dy = y - pos_->y();
  pos_->set(x, y);
  pos2_->set(pos2_->x() + dx, pos2_->y() + dy);
}

void GLineShape::moveImpl(float x, float y, bool callcb) {
  updateFrame();
  if (frame_.x == x && frame_.y == y) return;

  if (callcb && (eventmask_ & GEventType::Move) != 0) {
    GEvent e(GTrigger::move, GCore::elapsedTime());
    doFire(e);
  }
  if (layout_.passivePos) pos_->set(frame_.x, frame_.y);
}

// called when packing gadgets in a box, updates frame_ accordingly
void GLineShape::packPos(float parentx, float parenty) {
  auto initx = frame_.x, inity = frame_.y;
  /*
  // frame(x,y) must be the upper leftmost point
  auto x = min(pos_->x(), pos2_->x());
  auto y = min(pos_->y(), pos2_->y());
  frame_.x = parentx + x;
  frame_.y = parenty + y;
   */
  updateFrame();

  if ((eventmask_ & GEventType::Move) != 0
      && (frame_.x != initx || frame_.y != inity)) {
    GEvent e(on::move, GCore::elapsedTime());
    doFire(e);
  }
}

// finds the top-left position and updates the gadget frame
void GLineShape::updateFrame() {
  auto* par = getParent();
  if (!par || !pos_ || !pos2_) return;

  auto p1 = pos_->value(), p2 = pos2_->value();
  float x, y, w, h;

  if (p1.x <= p2.x) {x = p1.x; w = p2.x - p1.x + 1;}
  else {x = p2.x; w = p1.x - p2.x + 1;}

  if (p1.y <= p2.y) {y = p1.y; h = p2.y - p1.y + 1;}
  else {y = p2.y; h = p1.y - p2.y + 1;}

  frame_.x = par->frame_.x + x;
  frame_.y = par->frame_.y + y;
  frame_.w = w;
  frame_.h = h;
  //cerr << "GLineShape::updateFrame " << frame_<< " p1 " << p1 << " p1 " << p2 << endl;
  
  if (size_ && (size_->w() != w || size_->h() != h)) {
    size_->set(w, h, GSize::Pixels, GSize::Pixels);
  }
  update();
}

GHandle GLineShape::handleDown(GPoint const& wpos, bool inter) {
  if (!inter) return findHandle(wpos);
  else {
    *pos2_ = *pos_;
    return GHandle::Point2;
  }
}

void GLineShape::handleUp(GPoint const&, GHandle const& loc) {
  //updateLayout();
  if (auto* p = pos_->to<GRelativePos>()) p->updatePos();
  if (auto* p = pos2_->to<GRelativePos>()) p->updatePos();
}

void GLineShape::handleDrag(float dx, float dy, GHandle const& loc) {
  switch (loc) {
    case GHandle::Inside:
      //keepInParent(*par, p, w(), h());
      pos_->set(pos_->x() + dx, pos_->y() + dy);
      pos2_->set(pos2_->x() + dx, pos2_->y() + dy);
      return;

    case GHandle::Point1:
      pos_->set(pos_->x() + dx, pos_->y() + dy);
      return;

    case GHandle::Point2:
      pos2_->set(pos2_->x() + dx, pos2_->y() + dy);
      return;
      
    default:
      return;
  }
}

GHandle GLineShape::findHandle(GPoint const& wpos) {
  auto* par = getParent();
  if (!par) return GHandle::Outside;

  if (pickHandle(wpos, pos_->x() + par->frame_.x, pos_->y() + par->frame_.y)) {
    return GHandle::Point1;
  }
  else if (pickHandle(wpos, pos2_->x() + par->frame_.x, pos2_->y() + par->frame_.y)) {
    return GHandle::Point2;
  }
  else return GHandle::Inside;
}

Gadget* GLineShape::doPick(GPoint const& p, GPickMode& pm) {
  if (!modes_.shown || frame_.w <= 0.f || frame_.h <= 0.f) return nullptr;

  if (modes_.ignoreEvents && !pm.find_) {
    // callback?
    return nullptr;
  }
  
  auto* par = getParent();
  if (!par) return nullptr;

  GPoint a(pos_->x() + par->frame_.x, pos_->y() + par->frame_.y);
  GPoint b(pos2_->x() + par->frame_.x, pos2_->y() + par->frame_.y);
  GRect frame(a,b);

  // on teste sur zone agrandie car line est filiphorme
  auto handle = GPrefs::resizeHandleSize();
  frame.x -= handle/2;
  frame.y -= handle/2;
  frame.w += handle;
  frame.h += handle;

  if (!frame.contain(p)) return nullptr;

  //moved
  //if (pm.pred_ && !(pm.pred_)(*this)) return nullptr;

  GRect rect(p.x-handle/2, p.y-handle/2, handle, handle);
  if (!rect.intersect(a,b)) return nullptr;

  if (pm.pred_ && !(pm.pred_)(*this)) return nullptr;

  return this;
}

void GLineShape::doPaint(GRenderGraphics& g, GRender* parr) {
  auto* par = getParent();
  if (!par) return;

  GRender r(*this, parr);
  if (!r.paintGadget(g, GRender::DrawNone)) return;   // DrawMargin

  GPoint a(pos_->x() + par->frame_.x, pos_->y() + par->frame_.y);
  GPoint b(pos2_->x() + par->frame_.x, pos2_->y() + par->frame_.y);
  //cerr << "doPaint: " << a << " - " << b<< " / " << frame_<< endl;

  //if (!g.setClipRect(*this)) return;
  g.setColor(r.color_);
  g.unsetBgColor();
  g.drawLine(a,b);

  //if (isChecked() && layout_.hasSizeFrame) {
  if (isChecked() && (layout_.reshapeFrame & layout_.activeSize)) {
    auto handle = GPrefs::resizeHandleSize();
    g.cliprect_.x -= handle/2;
    g.cliprect_.y -= handle/2;
    g.cliprect_.w += handle;
    g.cliprect_.h += handle;
    g.setWinClip(g.cliprect_);
    g.setColor(GPrefs::resizeFrameColor());
    g.setBgColor(GPrefs::resizeHandleColor());
    paintHandle(g, a.x, a.y, handle);
    paintHandle(g, b.x, b.y, handle);
  }
}

void GLineShape::readAux(std::istream& in, GIOBuffer& buf) {
  if (!buf.read() || buf.tabCount() != buf.level()) return;
  if (buf.data().compare(0, 6, "topos: ") == 0) {
    *pos2_ = buf.data().substr(6);
    buf.clear();
  }
}

void GLineShape::writeAux(std::ostream& out, class GIOBuffer& buf) const {
  buf.printTabbed("topos: ");
  out << *pos2_ << '\n';
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GFreehandShape::GFreehandShape(GadgetType& t, GString const& spec) :
GShape(t, spec) {
  add(new GPos());
  add(new GSize());
}

GFreehandShape::GFreehandShape(GString const& spec) :
GFreehandShape(Type(), spec) {}

GFreehandShape& GFreehandShape::clone(GClone const& cm) {
  auto* g = new GFreehandShape;
  g->initClone(*this, cm);
  g->pmin_ = pmin_;
  g->pmax_ = pmax_;
  g->points_ = points_;
  return *g;
}

/*
 GBool& closed();
 void close(bool val)
GBool& GFreehandShape::closed() {return *closed_;}
void GFreehandShape::close(bool val) {*closed_ = val;}
*/

void GFreehandShape::addPoint(GPoint const& pt) {
  //if (points_.empty()) points_.push_back(GPoint(0,0));
  points_.push_back(pt);
  last_ = pt;

  if (pt.x < pmin_.x) pmin_.x = pt.x;
  else if (pt.x > pmax_.x) pmax_.x = pt.x;

  if (pt.y < pmin_.y) pmin_.y = pt.y;
  else if (pt.y > pmax_.y) pmax_.y = pt.y;

  float w = pmax_.x - pmin_.x + 1;
  float h = pmax_.y - pmin_.y + 1;
  //frame_.w = pmax_.x - pmin_.x + 1;
  //frame_.h = pmax_.y - pmin_.y + 1;
  size_->set(w, h, GSize::Pixels, GSize::Pixels);

  auto* par = getParent();
  if (!par) return;
  frame_.x = par->frame_.x + pos_->x() + pmin_.x;
  frame_.y = par->frame_.y + pos_->y() + pmin_.y;
  update();
  //cerr << "addPoint " << frame_ << endl;
}

void GFreehandShape::getPoints(GPoints& points) const {
  for (auto& it : points_) {
    points.push_back(GPoint(it.x - pmin_.x, it.y - pmin_.y));
  }
}

void GFreehandShape::moveImpl(float x, float y, bool callcb) {
  frame_.x = x + pmin_.x;
  frame_.y = y + pmin_.y;

  if (callcb && (eventmask_ & GEventType::Move) != 0) {
    GEvent e(GTrigger::move, GCore::elapsedTime());
    doFire(e);
  }
  if (layout_.passivePos) pos_->set(frame_.x, frame_.y);
}

// called when packing gadgets in a box, updates frame_ accordingly
void GFreehandShape::packPos(float parentx, float parenty) {
  auto initx = frame_.x, inity = frame_.y;
  frame_.x = parentx + pos_->x() + pmin_.x;
  frame_.y = parenty + pos_->y() + pmin_.y;

  if ((eventmask_ & GEventType::Move) != 0
      && (frame_.x != initx || frame_.y != inity)) {
    GEvent e(on::move, GCore::elapsedTime());
    doFire(e);
  }
}

GHandle GFreehandShape::handleDown(GPoint const& wpos, bool inter) {
  if (!inter) return GHandle::Inside;
  else return GHandle::Point2;
}

void GFreehandShape::handleUp(GPoint const&, GHandle const& loc) {
  //updateLayout();
}

void GFreehandShape::handleDrag(float dx, float dy, GHandle const& loc) {
  switch (loc) {
    case GHandle::Inside:
      pos_->set(pos_->x() + dx, pos_->y() + dy);
      return;

    case GHandle::Point2:
      addPoint(last_.x + dx, last_.y + dy);
      return;

    default:
      return;
  }
}

void GFreehandShape::doPaint(GRenderGraphics& g, GRender* parr) {
  auto* par = getParent();
  if (!par || points_.empty()) return;

  GRender r(*this, parr);
  if (!r.paintGadget(g, GRender::DrawNone)) return;   // DrawMargin

  if (!g.setClipRect(*this)) return;

  g.setColor(r.color_);
  //if (*closed_) g.setBgColor(r.background_); else g.unsetBgColor();
  g.unsetBgColor();

  g.translate(par->frame_.x + pos_->x(), par->frame_.y + pos_->y());

  //if (*closed_) g.drawPolygon(points_); else g.drawPolyline(points_);
  g.drawPolyline(points_);

  g.translate(-par->frame_.x - pos_->x(), -par->frame_.y - pos_->y());

  if (isChecked() && layout_.reshapeFrame) {
    paintHandleFrame(g);
  }
}

void GFreehandShape::readAux(std::istream& in, GIOBuffer& buf) {
  if (!buf.read() || buf.tabCount() != buf.level()) return;
  /*
  if (buf.data().compare(0, 8, "closed: ") == 0) {
    *closed_ = buf.data().substr(8);
    buf.clear();
  }*/

  if (!buf.read() || buf.tabCount() != buf.level()) return;
  if (buf.data().compare(0, 9, "points: [") == 0) {
    points_.clear();
    std::istringstream ss(buf.data().substr(9));
    buf.clear();
    
    while (ss) {
      GPoint pt;
      char sep;
      ss >> pt >> sep;
      //cout << "point: " << pt << " / " << sep << endl;
      addPoint(pt);
      if (sep == ']') break;
    }
  }
}

void GFreehandShape::writeAux(std::ostream& out, class GIOBuffer& buf) const {
  //buf.printTabbed("closed: "); out << *closed_ << "\n";
  buf.printTabbed("points: [");
  bool first = true;
  for (auto& it : points_) {
    if (first) first = false; else out << ", ";
    out << it;
  }
  out << "]\n";
}

}
