//
//  Table gadget
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <algorithm>
#include <giobuffer.hpp>
#include <gevent.hpp>
#include <gtrigger.hpp>
#include <gtable.hpp>
#include <gvalueitem.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
#include <gstyle.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_GADGET(Table)
GIMPLEMENT_GADGET(TRow)
GIMPLEMENT_GADGET(THead)

GTable::~GTable() {
  if (views_) {
    auto it = std::find(views_->begin(), views_->end(), this);
    if (it != views_->end()) views_->erase(it);
  }
}

GTable::GTable(GadgetType& t, GString const& spec)
: GBox(t, spec) {}

GTable::GTable(GString const& spec)
: GBox(Type(), spec) {}

GTable& GTable::clone(GClone const& clone) {
  auto* g = new GTable();
  g->initClone(*this, clone);
  return *g;
}

Gadget& GTable::createDefaultItem(GString const& arg) {
  return *new GTRow(arg);
}

int GTable::columnCount(int* minColumnCount) const {
  int maxcount = 0, mincount = std::numeric_limits<int>::max();
  for (auto& ch : children()) {
    if (auto* row = ch->toBox()) {
      int c = row->childCount();
      if (c > maxcount) maxcount = c;
      if (c < mincount) mincount = c;
    }
    else {
      if (1 > maxcount) maxcount = 1;
      if (1 < mincount) mincount = 1;
    }
  }
  if (minColumnCount) {
    *minColumnCount = (mincount==std::numeric_limits<int>::max()) ? 0 : mincount;
  }
  return maxcount;
}

Gadget* GTable::childAt(int row, int column) {
  if (row < 0 || column < 0) return nullptr;
  Gadget* r = childAt(row);
  if (!r) return nullptr;
  if (auto* rowbox = r->toBox()) return rowbox->childAt(column);
  else return r;
}

void GTable::perform(std::function<void(Gadget&)> fun) {
  for (auto& it : children()) {
    if (auto* row = it->toBox()) {
      if (!row->toTHead()) {
        for (auto& cell : row->children()) (fun)(*cell);
      }
    }
  }
}

void GTable::perform(std::function<void(Gadget&,int row, int column)> fun) {
  int r = 0;
  for (auto& it : children()) {
    if (auto* row = it->toBox()) {
      if (!row->toTHead()) {
        int c = 0;
        for (auto& cell : row->children()) (fun)(*cell, r, c++);
      }
    }
    ++r;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GTRow::GTRow(GadgetType& t, GString const& spec)
: GBox(t, spec) {}

GTRow::GTRow(GString const& spec)
: GBox(Type(), spec) {}

GTHead::GTHead(GadgetType& t, GString const& spec)
: GTRow(t, spec) {}

GTHead::GTHead(GString const& spec)
: GTRow(Type(), spec) {}

GTRow& GTRow::clone(GClone const& clone) {
  auto* g = new GTRow;
  g->initClone(*this, clone);
  return *g;
}

GTHead& GTHead::clone(GClone const& clone) {
  auto* g = new GTHead;
  g->initClone(*this, clone);
  return *g;
}

Gadget& GTRow::createDefaultItem(GString const& arg) {
  return *new GLabel(arg);
}

void GTRow::writeValues(std::ostream& out, GIOBuffer& buf) const {
  buf.setInsideRow(true);
  Gadget* last = childAt(-1);

  for (auto& it : children()) {
    if (it->hasValue()) {
      if (it.get() == last) buf.setInsideRow(false);
      it->writeValues(out, buf);
    }
  }
  buf.setInsideRow(false);
}

void GTRow::readValues(std::istream& in, GIOBuffer& buf) {
  buf.setInsideRow(true);
  Gadget* last = childAt(-1);
  if (toTHead()) return; // headers dont have values

  for (auto& it : children()) {
    if (!it->hasValue()) continue;
    if (it.get() == last) buf.setInsideRow(false);
    it->readValues(in, buf);
    if (!buf.read()) break;
  }
  buf.setInsideRow(false);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GDim GTRow::computeInsideSize(GRender& r) {
  GDim size;
  auto spacing2 = 0.f;     // horizontal spacing ingerited from Table
  if (r.parr_) spacing2 = r.parr_->gap_.crossspacing;

  bool one_found = false;
  float ch_h = 0.f;
  
  for (auto& ch : children_) {
    if (ch->isShown()) {
      ch->computeSize(&r, true);
      if (!ch->layout_.activePos && !ch->isTopLevel()) {
        ch_h = std::max(ch_h, ch->h());
        size.w += ch->w() + spacing2;
        one_found = true;
      }
    }
  }
  size.h += ch_h;
  if (one_found) size.w -= spacing2;
  return size;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GTable::pack(GRender* parr) {
  GRender r(*this, parr);
  GRect ch_area = childrenArea(r);
  packBorders(ch_area, r);  // changes ch_area!

  struct Column {
    float w{0.f};
    bool flex{false};
  };
  std::vector<Column> columns;
  size_t col_max = 0;
  
  packColumn(ch_area, r);
  
  for (auto& row_it : children_) {
    auto* row = row_it->toBox();
    if (!row) continue;
    
    row->modes_.dir = GFlow::Row;
    size_t col_nbr = 0;
    col_max = std::max(col_max, row->children_.size());
    columns.resize(col_max);
    
    for (auto& ch : row->children_) {
      if (ch->isShown() && !ch->isTopLevel()) {
        
        if (ch->layout_.activePos) {
          auto initx = ch->frame_.x, inity = ch->frame_.y;
          ch->pack(&r);
          if ((ch->eventmask_ & GEventType::Move) != 0
              && (ch->frame_.x != initx || ch->frame_.y != inity))
          {
            GEvent e(on::move, GCore::elapsedTime());
            doFire(e);
          }
          continue;
        }
        
        Column& c = columns[col_nbr];
        c.w = std::max(c.w, ch->frame_.w);
        if (ch->rowFlex(row)) c.flex = true;
        col_nbr++;
      }
    }
  }

  float h_used = 0.f;
  int flex_count = 0;
  
  for (auto& c : columns) {
    h_used += c.w;
    flex_count += c.flex;
  }

  float h_remain = ch_area.w - h_used;
  if (flex_count > 0) h_remain = h_remain / flex_count;
  
  auto spacing2 = r.gap_.crossspacing;  //  horizontal spacing

  for (auto& row_it : children_) {
    auto* row = row_it->toBox();
    if (!row) continue;
    
    size_t col_nbr = 0;
    float ch_x = ch_area.x;
    
    for (auto& ch : row->children_) {
      if (ch->isShown() && !ch->isTopLevel()) {
        
        GRect& chframe = ch->frame_;  // ne tient ps compte des Align !!!!
        GRect chframe_init = chframe;
        
        if (ch->layout_.activePos) {
          ch->pack(&r);
          ch->movedResizedCB(chframe_init);
          continue;
        }
        
        Column& c = columns[col_nbr];
        float ch_w = c.w;
        if (c.flex) ch_w += h_remain;
        chframe.x = ch_x;
        chframe.w = ch_w;
        ch_x += ch_w + spacing2;
        ch->pack(&r);
        col_nbr++;
        ch->movedResizedCB(chframe_init);
      }
    }
  }
}

}
