//
//  GUI events, Triggers, Callbacks and Notifiers
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
#include <algorithm>
#include <gapp.hpp>
#include <gevent.hpp>
#include <gtrigger.hpp>
#include <gbox.hpp>
#include <gchoice.hpp>
#include <gexpr.hpp>
#include <gcontrol.hpp>
#include <gstring.hpp>
using namespace std;
namespace guit {

GTrigger
GTrigger::none(GEventType::NoEvent, -1),
GTrigger::changeaction(GEventType::Action | GEventType::Change, -1),
GTrigger::action(GEventType::Action, -1),
GTrigger::done(GEventType::Done, 0),
GTrigger::changing(GEventType::Changing, -1),
GTrigger::change(GEventType::Change, -1),
GTrigger::checked(GEventType::Change, 1),
GTrigger::unchecked(GEventType::Change, 0),
GTrigger::click(GEventType::Click, 1),
GTrigger::dblclick(GEventType::Click, 2),
GTrigger::mousedown(GEventType::MouseButton, 1),
GTrigger::mouseup(GEventType::MouseButton, 0),
GTrigger::mousemove(GEventType::MouseMove, -1),
GTrigger::mousedrag(GEventType::MouseDrag, -1),
GTrigger::mouseover(GEventType::MouseCross, 1),
GTrigger::mouseout(GEventType::MouseCross, 0),
GTrigger::wheel(GEventType::Wheel, 1),
GTrigger::keydown(GEventType::Key, 1),
GTrigger::keyup(GEventType::Key, 0),
GTrigger::keytype(GEventType::Key, 2),
GTrigger::caret(GEventType::Caret, -1),
GTrigger::touch(GEventType::Touch, -1),
GTrigger::gesture(GEventType::Gesture, -1),
GTrigger::focus(GEventType::Focus, 1),
GTrigger::focusout(GEventType::Focus, 0),
GTrigger::enabled(GEventType::Enable, 1),
GTrigger::disabled(GEventType::Enable, 0),
GTrigger::show(GEventType::Show, 1),
GTrigger::hide(GEventType::Show, 0),
GTrigger::move(GEventType::Move, -1),
GTrigger::paint(GEventType::Paint, -1),
GTrigger::resize(GEventType::Resize, -1),
GTrigger::init(GEventType::Init, 1),
GTrigger::add(GEventType::AddRemove, 1),
GTrigger::remove(GEventType::AddRemove, 0),
GTrigger::cut(GEventType::CutPaste, 0),
GTrigger::copy(GEventType::CutPaste, 1),
GTrigger::paste(GEventType::CutPaste, 2),
GTrigger::user(GEventType::User, 1);

void GTrigger::receiverAdded(Gadget& g, GCond& p) {
  g.eventmask_ = g.eventmask_ | mask_;
}

void GTrigger::receiverRemoved() {
  // update eventmask !!!!                   // TODO !!!!
}

bool GTrigger::operator==(GTrigger const& other) const {
  //return mask_ == other.mask_ && filter_ == other.filter_;
  return
  isEventTrigger() && other.isEventTrigger()
  && mask_ == other.mask_ && filter_ == other.filter_;
}

GTrigger* GTrigger::getTrigger(GString const& trigger) {
  static std::map<GString, GTrigger*> triggers_;

  if (triggers_.empty()) {
    triggers_ = {
      {"none", &GTrigger::none},
      {"action", &GTrigger::action},
      {"done", &GTrigger::done},
      {"change", &GTrigger::change},
      {"changing", &GTrigger::changing},
      {"checked", &GTrigger::checked},
      {"unchecked", &GTrigger::unchecked},
      {"click", &GTrigger::click},
      {"dblclick", &GTrigger::dblclick},
      {"mousedown", &GTrigger::mousedown},
      {"mouseup", &GTrigger::mouseup},
      {"mousedrag", &GTrigger::mousedrag},
      {"mousemove", &GTrigger::mousemove},
      {"mouseover", &GTrigger::mouseover},
      {"mouseout", &GTrigger::mouseout},
      {"wheel", &GTrigger::wheel},
      {"keydown", &GTrigger::keydown},
      {"keyup", &GTrigger::keyup},
      {"keytype", &GTrigger::keytype},
      {"caret", &GTrigger::caret},
      {"touch", &GTrigger::touch},
      {"gesture", &GTrigger::gesture},
      {"focus", &GTrigger::focus},
      {"focusout", &GTrigger::focusout},
      {"enabled", &GTrigger::enabled},
      {"disabled", &GTrigger::disabled},
      {"show", &GTrigger::show},
      {"hide", &GTrigger::hide},
      {"move", &GTrigger::move},
      {"init", &GTrigger::init},
      {"resize", &GTrigger::resize},
      {"paint", &GTrigger::paint},
      {"add", &GTrigger::add},
      {"remove", &GTrigger::remove},
      {"copy", &GTrigger::copy},
      {"cut", &GTrigger::cut},
      {"paste", &GTrigger::paste},
      {"user", &GTrigger::user},
    };
  }
  auto it = triggers_.find(trigger);
  return it == triggers_.end() ? nullptr : it->second;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GEvent::GEvent(GTrigger const& trigger, GTime time)
: trigger_(&trigger), time_(time) {}

GEvent::~GEvent() {
  // fired at the end if not null
  if (postfire_) (postfire_)(*this);
}

void GEvent::setMouseButtonHandler(int chan, std::function<bool(GMouseEvent&)> cb) {
  GCore::obtainChannel(chan).buttonhandler_ = cb;
}

void GEvent::setMouseMotionHandler(int chan, std::function<bool(GMouseEvent&)> cb) {
  GCore::obtainChannel(chan).movehandler_ = cb;
}

void GEvent::setWheelHandler(int chan, std::function<bool(GMouseEvent&)> cb) {
  GCore::obtainChannel(chan).wheelhandler_ = cb;
}

void GEvent::setKeyHandler(int chan, std::function<bool(GKeyEvent&)> cb) {
  GCore::obtainChannel(chan).keyhandler_ = cb;
}

GMouseEvent& GEvent::asMouseEvent() {
  throw GFatalError("asMouseEvent","Event can't be converted to GMouseEvent");
}

GKeyEvent& GEvent::asKeyEvent() {
  throw GFatalError("asKeyEvent","Event can't be converted to GKeyEvent");
}

GTouchEvent& GEvent::asTouchEvent() {
  throw GFatalError("asTouchEvent","Event can't be converted to GTouchEvent");
}

GGestureEvent& GEvent::asGestureEvent() {
  throw GFatalError("asGestureEvent","Event can't be converted to GGestureEvent");
}


// NOTE: defined in the native file gsdl.cpp
// GEventState GEvent::keyToName, nameToKey, modifierToState

GEventState GEvent::buttonToState(int mouse_button) {
  switch (mouse_button) {
    case 1: return GEventState::Button1;
    case 2: return GEventState::Button2;
    case 3: return GEventState::Button3;
    case 4: return GEventState::Button4;
    case 5: return GEventState::Button5;
    default: return GEventState::NoState;
  }
}

static GEventState modifierNameToState(GString const& mod) {
  switch (strid(strlower(mod))) {
    case strid("⌃"):
    case strid("ctrl"):
      return GEventState::Ctrl;
    case strid("⇧"):
    case strid("shift"):
      return GEventState::Shift;
    case strid("⌘"):
    case strid("cmd"):
    case strid("meta"):
      return GEventState::Meta;
    case strid("⌥"):
    case strid("alt"):
      return GEventState::Alt;
    case strid("altgr"):
      return GEventState::AltGr;
    case strid("caps"):
      return GEventState::Caps;
    case strid("num"):
      return GEventState::Num;
    default:
      return GEventState::NoState;
  }
}

GEventState GEvent::modifiersToStates(GString const& modifiers) {
  GEventState state{};
  GStrings v;
  strsplit(modifiers, v, "-+");
  for (auto& it : v) state = state | modifierNameToState(it);
  return state;
}

//⌦⌧⌫↩︎⏎⇥⇤⌤␣⌃↑←↓→⚙︎⌂⌗⎇✆☎︎☏♲⏣⌬⌛︎
/* SDL sait pas les afficher!
GString GEvent::toModifierSymbols(GEventState state) {
  GString str;
  // repects apple order on menus
  if (state & GEvent::CtrlMode) str += "⌃";
  if (state & GEvent::AltMode) str += "⌥";
  if (state & GEvent::ShiftMode) str += "⇧";
  if (state & GEvent::MetaMode) str += "⌘";
  if (state & GEvent::AltGrMode) str += "⎇";
  if (state & GEvent::CapsMode) str += "⇪";
  if (state & GEvent::NumMode) str += "⌧";
  return str;
}*/


int GEvent::matches(GTrigger const& tr, int channel) const {
  if (auto* t = tr.toExprTrigger()) {
    if (!gadget_) {
      GApp::error("GEvent::matches", "Internal error: null sender ");
      return 0;
    }
    if (!t->recv_ || t->invalid_) return 0;
    
    // sinon un GEvent type ou une autre cause pourrait activer le callback
    auto it = std::find(t->senders_.begin(), t->senders_.end(), gadget_);
    if (it == t->senders_.end()) return 0;
    
    if (!t->boolexpr_) return 0;
    else if (t->boolexpr_->boolValue()) return 1;
    else return -1;
  }
  else {
    if ((tr.mask_ & trigger_->mask_) != 0 && (channel < 0 || channel == channel_)) {
      if (tr.filter_ < 0 || tr.filter_ == trigger_->filter_) return 1;
      else return -1;
    }
    else return 0;
  }
}

GEventType GEvent::eventMask() const {
  return trigger_ ? trigger_->mask_ : GEventType::NoEvent;
}

GWindow* GEvent::window() const {
  if (window_) return window_;
  else if (gadget_) return gadget_->getWindow();
  else return nullptr;
}

Gadget& GEvent::from() const {
  if (gadget_) return *gadget_;
  else throw GFatalError("GEvent::gadget","null gadget");
}

GBox& GEvent::fromBox() const {
  if (!gadget_) throw GFatalError("GEvent::box","null gadget");
  else if (!gadget_->toBox()) throw GFatalError("GEvent::box","gadget is not a GBox");
  return *gadget_->toBox();
}

GProp& GEvent::fromProp() const {
  if (!aux_ || !aux_->toProp()) throw GFatalError("GEvent::prop","null prop");
  else return *aux_->toProp();
}

/*
GTextData* GEvent::data() const {
  if (!gadget_) return nullptr;
  if (auto* p = gadget_->findProp<GTextData>()) return p;
  return nullptr;
}*/

Gadget* GEvent::chosen() const {
  if (!aux_ || !gadget_ || !gadget_->toBox()) return nullptr;
  return aux_->toGadget();
}

/*
int GEvent::boxIndex() const {
  GBox* box{};
  if (!aux_ || !gadget_ || !(box = gadget_->toBox())) return -1;

  // necessaire pour les combos
  if (auto* choice = box->choiceProp()) return choice->chosenIndex();

  // pas correct car devrait etre delegue, c'est a la box de le faire
  int n = 0;
  for (auto& it : box->children()) {
    if (it.get() == aux_) return n;
    else if (it->isActivable()) n++;
  }
  return -1;
}*/

void GEvent::postFire(std::function<void(GEvent&)> const& todo) {
  postfire_ = todo;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GMouseEvent::GMouseEvent(GTrigger& t, GMouseEvent const& from)
: GMouseEvent(from) {
  trigger_ = &t;
  postfire_ = nullptr;
}

GPoint GMouseEvent::pos() const {
  if (gadget_)
    return GPoint(wpos_.x - gadget_->wx(), wpos_.y - gadget_->wy());
  else
    return GPoint();
}

void GMouseEvent::redirectEvent(Gadget& g) {
  auto* c = GCore::channel(channel_);
  if (!c) return;

  if (c->pressg_ && c->pressg_->isActivable()) {
    c->pressg_->modes_.active = false;
  }
  c->pressg_ = g;
  gadget_ = &g;
  //cerr << "redirect: " << c->pressg_ << endl;
  g.doFire(*this);
}

}
