//
// files and file paths
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#if defined(_WIN32)
#  include <direct.h>
#endif
#include <algorithm>
#include <fstream>
#include <cstring>
#include <cstdio>
#include <sys/types.h>
#include <sys/stat.h>
#include <gstring.hpp>
#include <gfile.hpp>
#include <gapp.hpp>
#include <gcontrol.hpp>
#include <gstring.hpp>
namespace guit {

bool openFile(std::ifstream& f, GString const& path, const char* locale) {
  if (path.empty()) return false;

  static GString jokers("./\\~:!");

  if (jokers.find(path[0]) != GString::npos) {
    f.open(normPath(path));
  }
  else {
    auto p = normPath(path);
    f.open(p);

    if (!f.is_open()) {
      for (auto& it : GApp::resourceDirs()) {
        f.open(it + p);
        if (f.is_open()) break;
      }
    }
  }

  if (!f.is_open()) return false;
  if (locale) f.imbue(std::locale(locale));
  return true;
}

bool openFile(std::ofstream& f, GString const& path, const char* locale) {
  if (path.empty()) return false;
  f.open(normPath(path));

  if (!f) return false;
  if (locale) f.imbue(std::locale(locale));
  return true;
}

/*
GString getenvvar(const char* var) {
  GString result;
#if defined(_WIN32)
  size_t len{};
  char value[201];
  if (getenv_s(&len, value, 200, var) == 0 && len != 0) result = value;
#else
  // global resource, dont free it!
  if (char* value = ::getenv(var)) result = value;
#endif
  // check if / terminated
  if (!result.empty() && result.back() != GPathSepar) result += GPathSepar;
  return result;
}*/


static GString getHome() {
  static GString dir;
  static bool init{false};

  if (!init) {
    init = true;
#if defined(_WIN32)
    size_t len{};
    char value[201];
    if (getenv_s(&len, value, 200, "USERPROFILE") == 0 && len != 0) dir = normPath(value);
#else
    // global resource, dont free it!
    if (const char* value = ::getenv("HOME")) dir = normPath(value);
#endif
    if (dir.empty()) gerror("normpath","home directory could not be found");
  }

  return dir;
}


GString normPath(GString const& path, bool add_slash) {
  if (path.empty()) return "";

  GString p = path;
#if defined(_WIN32)
  for (auto& c : p) {if (c == '\\') c = '/';}
#endif
  switch (p[0]) {
    case ':':
      p.replace(0, 1, GApp::resourceDirs()[0]);
      break;
    case '!':
      p.replace(0, 1, GApp::appDir());
      break;
    case '~':
      if (!getHome().empty()) p.replace(0, 1, getHome());
      break;
  }

  if (add_slash && p.back() != '/') p += '/';
  return p;
}

GString parentPath(GString const& path) {
  GString s(path);
  size_t pos = s.find_last_of("/\\");                // FAUX si !!
  if (pos == GString::npos) s = "/"; else s.erase(pos);
  return s;
}

bool fileExist(GString const& path) {
  if (path.empty()) return false;
  struct stat fstat;
  return (::stat(normPath(path).c_str(), &fstat) == 0);
}

bool makeDir(GString const& path) {
  if (path.empty()) return false;
#if defined(_WIN32)
  return _mkdir(normPath(path).c_str()) == 0;
#else
  return ::mkdir(normPath(path).c_str(), S_IRWXU) == 0;
#endif
}

int system(GString const& command) {
  return ::system(command.c_str());
}

bool copyFile(GString const& from, GString const& to) {
  // note: <filesystem> provides standard functions in C++17
  std::ifstream src(from);
  std::ofstream dst(to);
  if (!src || !dst) return false;
  char c;
  while (src.get(c)) dst << c;
  return true;
}

GString basename(GString const& path, bool with_extension) {
  auto p = normPath(path);
  if (p.empty()) return ".";
  if (p.size() == 1) return p;
  if (p[p.size()-1] == '/') return basename(p.substr(0, p.size()-1));
  size_t pos = p.find_last_of('/');
  if (with_extension) return (pos == std::string::npos) ? p : p.substr(pos+1);
  else {
    ssize_t pos2 = p.size()-1, pos_min = (pos == std::string::npos) ? 0 : pos+1;
    while (pos2 >= pos_min && p[pos2] != '.') --pos2;
    if (pos2 < pos_min) return (pos == std::string::npos) ? p : p.substr(pos+1);
    else return (pos == std::string::npos) ? p.substr(0, pos2) : p.substr(pos+1, pos2-pos-1);
  }
}

GString dirname(GString const& path) {
  auto p = normPath(path);
  if (p.empty()) return ".";
  if (p.size() == 1) {
    return p[0] == '/' ? "/" : ".";
  }
  if (p[p.size()-1] == '/') return dirname(p.substr(0, p.size()-1));
  size_t pos = p.find_last_of('/');
  if (pos == std::string::npos) return ".";  // no slash
  return pos == 0 ? "/" : p.substr(0, pos);
}

GString extname(GString const& path, bool with_dot) {
  ssize_t pos = path.size()-1;
  while (pos >= 0 && path[pos] != '.' && path[pos] != '/' && path[pos] != '\\') --pos;
  if (pos < 0 || path[pos] == '/' || path[pos] == '\\') return "";  // no extension
  return with_dot ? path.substr(pos) : path.substr(pos+1);
}

}
