//
// Gadget Inspector
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <list>
#include <sstream>
#include <guit.hpp>
#include <gshape.hpp>
#include <gcursor.hpp>
#include <gcomputedstyle.hpp>
#include <ginspector.hpp>
#include <gcontrol.hpp>
using namespace std;
namespace guit {

class GGlass : public GShape {
public:
  GGlass(GString const& css = "");

  void highlight(Gadget* obj, bool state);
  void adaptTo(Gadget& obj);
};

GGlass::GGlass(GString const& css) : GShape(css) {
  setInternal();
  ignoreEvents(true);
  posProp().setLayer(GPos::MaxLayer + 1);
}

void GGlass::highlight(Gadget* obj, bool state) {
  if (state && obj) {
    if (auto* win = obj->getWindow()) {
      win->add(this);
      adaptTo(*obj);
    }
  }
  else {
    if (auto* par = this->getParent()) par->remove(this);
  }
}

void GGlass::adaptTo(Gadget& obj) {
  move(obj.wpos());
  resize(obj.dim());
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GInspector& GApp::inspector() {
  GInspector& i = GInspector::instance();
  GCore::inspector_ = &i;
  GCore::doinspect_ = [&i](Gadget* g){i.inspect(g);};
  return i;
}

void GApp::showInspector() {
  auto& i = inspector();
  i.show();
  i.raise();
}

GInspector& GInspector::instance() {
  static GInspector i;
  GCore::inspector_ = &i;
  return i;
}

GInspector::~GInspector() {}

GInspector::GInspector() :
GWindow("#inspector"),
targetglass_(new GGlass("#glass {background: 1 0.5 0 0.1}")),
highlightglass_(new GGlass("#glass {background: 1 0 0 0.2}"))
{
  GApp::addCssRules
  ("#inspector {pos:700 0; size:400 600; margin:8 3 8 3; text: Guit Inspector}"
   "#inspector #onoffbox {flex:1 0; margin:8; align:x x; gap:1}"
   "#inspector #onoffbox #buttons {flow:row; border:none; margin:0; font:bold}"
   "#inspector #targetbox {flex: 1 0; align:x x; gap:8 10; border:rounded 8 12;"
   "           background:white; margin:0 8 3 3}"
   "#inspector ScrollBox {background:white}"
   "#inspector #propbox {flex:1 0; align:x x; gap:8 5; margin:5}"
   "#inspector #childbox {flex:1 0; align:x x; gap:8 5; margin:5}"
   "#inspector .target {font:bold; color:white; background:guitblue}"
   "#inspector .target:active {color:guitblue; background:salmon}"
   "#inspector .child {font:bold}"
   "#inspector .child:active {background:salmon}"
   "#inspector .prop {font:bold; color:green}"
   "#inspector .text {color:brown}"
   "#inspector .select {text:>; font:bold; tip:Select Gadget}"
   "#inspector .delete {text:x; font:bold; color:firebrick; tip:Delete}"
   "#inspector .add {text:+; font:bold;}"
   "#inspector .undo {text:undo; pos: 5 5; background:green; color:white; font:bold}"
   "#inspector .undo:active {background:lightgreen}"
   "#inspector .instyle {text:in style; color:coral}"
   "#inspector .ingadget {text:in gadget; color:guitblue}"
   "#inspector .constprop {text:in gadget (const); color:navy}"
   );

  setInternal();

  inspectchoice_ << [this]{
    int choice = inspectchoice_();
    GCore::inspecting_ = choice;
    if (choice == 0) inspect(nullptr);
    GApp::updateAll();
    if (inspectingcb_) inspectingcb_(choice);
  };

  auto& onoffbox = HBox("#onoffbox")
  << "<b>Inspect: "
  << (ButtonBox("#buttons")
      << inspectchoice_ << " Off " << " Edit " << " Live "
      );

  auto& targetbox = Table("#targetbox")
  << (TRow
      << "<b>Target: "s
      << (LinkItem(".target")
          << hastarget_
          << targettype_
          << on::mousedown / [this]{highlight(target(),true);}
          << on::mouseup   / [this]{highlight(target(),false);}
          )
      << (Button(".delete")
          << hastarget_
          << [this]{deleteGadget(target(),true);}
          )
      << (Label(".text") << targetext_)
      )
  << (TRow
      << "<b>Parent:"s
      << (LinkItem(".child")
          << hasparent_
          << parenttype_
          << on::mousedown / [this]{highlight(targetparent_,true);}
          << on::mouseup   / [this]{highlight(targetparent_,false);}
          )
      << (Button(".select ^")
          << hasparent_
          << [this]{inspect(targetparent_);}
          )
      << (Label(".text") << parentext_)
      )
  << (Label << targetgeom_)
  ;

  auto& tabbox = TabBox
  << (propscroll_  << Title("Properties") << propbox_)
  << (childscroll_ << Title("Children") << childbox_)
  /*                                       BUG si mis avant !!!
   << (Button(".undo")
   << Pos(5,5)
   << hasundo_
   << [this]{undo();}
   )*/
  ;

  auto& undo = Button(".undo")
  << new GRelativePos(tabbox, 5,5)
  << hasundo_
  << [this]{undelete();};

  *this
  << on::hide / [this]{inspectchoice_ = 0;}
  << onoffbox
  << targetbox
  << tabbox
  << undo;
}

GInspector& GInspector::show(bool show) {
  GWindow::show(show);
  setInspecting(show);
  return *this;
}

void GInspector::highlight(Gadget* obj, bool state) {
  highlightglass_->highlight(obj, state);
}

int GInspector::inspecting() {
  return GCore::inspecting_;
}

void GInspector::setInspecting(int mode) {
  inspectchoice_ = mode;
  if (mode && !isShown()) show(true);   // !isShown() to avoid focus changes
  GApp::updateAll();
}

Gadget* GInspector::inspected() {
  return GCore::inspected_;
}

void GInspector::inspect(Gadget* obj) {
  if (!obj) {
    if (target()) {
      targetglass_->highlight(target(), false);
      if (inspectedcb_) inspectedcb_(nullptr);
    }
    GCore::inspected_ = nullptr;
    target_ = nullptr;
    targetparent_ = nullptr;
    hastarget_ = false;
    hasparent_ = false;
  }

  else {  // obj
    GWindow* win{};
    if (obj != target()
        && !obj->isInternal()
        && (win = obj->getWindow())
        && !win->isInternal()
        ) {
      GCore::inspected_ = obj;
      target_ = obj;
      hastarget_ = true;

      if (obj->getParent()) {
        targetparent_ = obj->getParent();
        hasparent_ = true;
      }
      else {
        targetparent_ = nullptr;
        hasparent_ = false;
      }
      targetglass_->highlight(obj, true);
      if (inspectedcb_) inspectedcb_(obj);
    }

    /* FOIREUX !è!!
     GCursor::cursor()->startDrag(obj,[this,&obj](GCursor& c){
     highlight(&obj, false);
     moveGadget(c, obj);
     setTarget(obj);
     updateValues();
     });
     */
  }
  updateValues();
}

void GInspector::setInspectingCB(std::function< void(int) > const& fun) {
  inspectingcb_ = fun;
}

void GInspector::setInspectedCB(std::function< void(Gadget*) > const& fun) {
  inspectedcb_ = fun;
}

void GInspector::updateValues() {
  if (target()) {
    targettype_ = target()->typeName();
    targetext_ =  target()->text();
    std::ostringstream ss;
    ss.imbue(std::locale("C"));
    ss.precision(3);
    auto* t = target();

    ss
    << "<b>Geometry:  </b>"
    << "Pos: " << t->x() << " " << t->y()
    << "  -  WPos: " << t->wx() << " " << t->wy()
    << "  -  Size: " << t->w() << " " << t->h()
    << "  -  Layer: " << t->zlayer()
    ;
    targetgeom_ = ss.str();
  }
  else {
    targettype_ = "";
    targetext_  = "";
    targetgeom_ = "";
  }

  if (targetparent_) {
    parenttype_ = targetparent_->typeName();
    parentext_  = targetparent_->text();
  }
  else {
    parenttype_ = "";
    parentext_  = "";
  }

  updateChildValues();
  updatePropValues();
  update();
}

void GInspector::updateChildValues() {
  childbox_.removeChildren();
  if (!target()) return;

  if (auto* box = target()->toBox()) {
    for (auto& c : box->children()) {
      if (!c->isInternal()) {
        childbox_
        << (TRow
            << (LinkItem(".child")
                << c->typeName()
                // << on::mousedown / [this,c]{moveTarget(c);}
                << on::mousedown / [this,c]{highlight(c,true);}
                << on::mouseup   / [this,c]{highlight(c,false);}
                )
            << (Button(".select") << [this,c]{inspect(c);})
            << (Button(".delete") << [this,c]{deleteGadget(c,false);})
            << (Label(".text") << c->text())
            );
      }
    }
  }
}

void GInspector::updatePropValues() {
  propbox_.removeChildren();
  if (!target()) return;

  GComputedStyle cs(*target());

  // doit etre preserve a cause des callbacks!
  static GPropMap* map{};
  if (map) delete map;
  map = new GPropMap(cs.getProps());

  for (auto& it : target()->props()) {
    if (it->role() > GPropType::Abstract) (*map)[it->role()] = it;
  }

  propbox_
  << (Label("{align:c c; margin:10}")
      // << "[+] adds prop to gadget - [x] removes it"  BUG! met en bold !
      << "<i> [+] adds prop to gadget - [x] removes it"
      );

  for (auto& it : *map) {
    GProp* p = it.second;

    auto& row = TRow();
    row << ( Label(".prop") << p->typeName() );

    if (!target()->contain(*p)) {   // comes from style
      row
      << ".instyle"
      << (Button(".add") << [this,p]{addProperty(target(), p);})
      << p->stringValue();
    }
    else if (auto* m = p->toMutableProp()) {
      auto& field = (TextField << p->stringValue());
      row
      << ".ingadget"
      << (Button(".delete") << [this,p]{deleteProperty(target(), p);})
      << (field << [m,&field]{ m->set(field.text());});
    }
    else {
      row
      << ".constprop"
      << (Button(".add") << [this,p]{addProperty(target(), p);})
      << p->stringValue();
    }
    propbox_ << row;
  }
}

void GInspector::deleteProperty(Gadget* parent, GProp* prop) {
  if (!parent || !prop) return;

  hasundo_ = true;
  addUndo(*prop, parent);
  parent->remove(prop);
  parent->updateLayout(Update::Now);
  updateValues();
}

void GInspector::addProperty(Gadget* parent, GProp* prop) {
  if (!parent || !prop) return;
  parent->add(prop->clone(true));
  parent->updateLayout(Update::Now);
  updateValues();
}

void GInspector::deleteGadget(Gadget* obj, bool itself) {
  if (!obj) return;
  GBox* parent = obj->getParent();
  if (!parent) {
    /// nope?
  }
  else {
    hasundo_ = true;
    addUndo(*obj, parent);
    parent->remove(obj);

    if (itself) inspect(nullptr);
    else if (parent == target()) {
      // update size of par glass, needs to update layout first
      parent->updateLayout(Update::Now);
      targetglass_->adaptTo(*parent);
      updateValues();
    }
  }
}

void GInspector::addGadget(GBox& parent, Gadget& obj, int ix) {
  parent.add(obj, ix);
  parent.updateLayout(Update::Now);
  updateValues();
}


void GInspector::addUndo(Gadget& obj, GBox* parent) {
  int pos = 0;
  if (parent) pos = parent->childPos(obj);
  undolist_.push_back(UndoElem(obj, parent, pos));
}

void GInspector::addUndo(GProp& obj, Gadget* parent) {
  undolist_.push_back(UndoElem(obj, parent));
}

void GInspector::undelete() {
  if (undolist_.empty()) return ;
  UndoElem elem = undolist_.back();
  undolist_.pop_back();
  hasundo_ = !undolist_.empty();

  Gadget * obj{};

  if (auto* p = elem.obj->toProp()) {
    if (elem.parent && (obj = elem.parent->toGadget())) {
      obj->add(p);
    }
  }
  else if (auto* g = elem.obj->toGadget()) {
    GBox* box{};
    if (elem.parent && (box = elem.parent->toBox())) {
      box->add(g, elem.pos);
      obj = box;
    }
  }

  if (obj) {
    obj->getWindow()->updateLayout(Update::Now);
    // update size of par glass, needs to update layout first
    if (obj == target()) targetglass_->adaptTo(*obj);
    updateValues();
  }
}

/*
void GInspector::moveTarget(Gadget* obj) {
  if (!obj) return;

  GWindow* win{};
  if (!obj->isInternal() && (win = obj->getWindow()) && !win->isInternal()) {
    highlight(obj, true);
    GCursor::cursor()->startDrag(childbox_,
                                 [this,obj](GCursor& c){moveTargetCB(c, *obj);});

  }
}

void GInspector::moveTargetCB(GCursor& c, Gadget& obj) {
  highlight(&obj, false);
  moveGadget(c, obj);
}
 */

bool GInspector::moveGadget(GCursor& c, Gadget& moved) {
  auto* win = c.targetWindow();
  if (!win) return false;

  auto* found = win->childAtWPoint(c.targetPos(), true/*rec*/);
  if (!found || found == &moved || moved.isInternal()) return false;

  if (found == &childscroll_ || childscroll_.contain(found, true)) {
    int ix = childbox_.nearestChildIndex(c.targetPos(), false);
    if (target() && target()->toBox()) {
      addGadget(*target()->toBox(), moved, ix);
      return true;
    }
  }
  else {
    if (win->isInternal()) return false;
    auto* box = found->toBox();
    if (!box) box = found->getParent();
    if (!box) return false;
    int ix = box->nearestChildIndex(c.targetPos(), false);
    addGadget(*box, moved, ix);
    return true;
  }
  return false;
}

}
