//
// Border Prop
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef GuitBorder_hpp
#define GuitBorder_hpp
#include <gmargin.hpp>
namespace guit {

/// @internal
class GBorderValue : public GMarginValue {
public:
  enum Kind {None=0, Squared=1<<0, Rounded=1<<2, Oval=1<<3, Polygon=1<<4, TextEdit=1<<5};

  ~GBorderValue();
  GBorderValue() = default;
  GBorderValue(Kind type, float left, float right, float top, float bottom,
               GColor& idle_color, GColor& active_color);
  GBorderValue(GPoints const& polygon, GColor& idle, GColor& selected);
  GBorderValue(GBorderValue const&);

  GBorderValue& operator=(GBorderValue const&);
  bool operator==(const GBorderValue&) const;
  bool operator!=(const GBorderValue&) const;
  friend std::ostream& operator<<(std::ostream& out, const GBorderValue&);

  bool hasDrawShape(Gadget&);
  bool hasPickShape(Gadget&);
  void drawBorder(GRenderGraphics&, GRender&);
  bool pickBorder(Gadget&, GRect const&, GPoint const&);

  Kind style{};
  gptr<GColor> idlecolor, activecolor;
  GPoints* polygon{};
};

inline bool gconvert(GBorderValue const& from, GBorderValue& to) {to = from; return true;}
bool gconvert(GString const& from, GBorderValue& to);
bool gconvert(GBorderValue const& from, GString& to);


/// Border and padding of a gadget.
/// Allows specifying:
/// - the border 'kind'
/// - the padding inside the border
/// - the idle and checked (ie selected) color of the border (the gadget color is used if unspecified)
/// - a polygonal border
///
/// @note Gadgets that contain editable text (e.g. GTextField) use GBorder::textedit which color changes
/// when the gadgets gets the focus. The focus wont be shown if another border kind is used.
class GBorder : public GVarProp_<GBorderValue> {
public:
  using Super = GVarProp_<GBorderValue>;
  enum Kind {None=0, Squared=1<<0, Rounded=1<<2, Oval=1<<3, Polygon=1<<4, TextEdit=1<<5};

  /// predefined borders.
  static GBorder none, squared, rounded, oval, polygon, textedit;

  GBorder();

  /// constr: value = "kind  left_right_padding  top_bottom_padding  idle_color  selected_color
  /// examples:
  /// - border:rounded
  /// - border:rounded 2 2
  /// - border:rounded 2 2 gray blue
  /// - border:rounded 2 2 '0.7 0.7 0.7' blue
  GBorder(GString const& value);

  /// uses gadgets' Color prop if _color_ is null.
  GBorder(Kind, float left_right_padding, float top_bottom_padding,
          GColor& idle_color, GColor& checked_color);

  GBorder(GPoints const& polygon, GColor& idle_color, GColor& active_color);

  GBorder(GBorderValue const&);

  GBorder(GBorder const& from, float left_right_padding, float top_bottom_padding,
          GColor* idle_color = {}, GColor* active_color = {});

  GBorder* clone(bool copy_value) const override;

  using GType = GPropType_<GBorder,GProp>;
  static GType& Type();
  GType& type() const override {return Type();}

protected:
  friend class GRender;
  void apply(Gadget*, GRender&, Specif) final;
  void updateGadget(Gadget&) override;
};


/// Return a GBorder created in the heap.
/// @{
inline GBorder& Border() {return *new GBorder();}
inline GBorder& Border(GString const& value) {return *new GBorder(value);}
/// @}

}
#endif
