//
// Prop that specifies whether a gadget is flexible horinzontaly ans vertically
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef GuitFlex_hpp
#define GuitFlex_hpp
#include <gvarprop.hpp>
namespace guit {

class GFlexValue : public GDim {
public:
  GFlexValue() = default;
  GFlexValue(int8_t flex1, int8_t flex2, int8_t mode);
  bool operator==(GFlexValue const&) const;
  bool operator!=(GFlexValue const&) const;
  int8_t flex1{}, flex2{}, mode{};
};

inline bool gconvert(GFlexValue const& from, GFlexValue& to) {to = from; return true;}
bool gconvert(GString const& from, GFlexValue& to);
bool gconvert(GFlexValue const& from, GString& to);


/**  Specifies whether a gadget is flexible.
 * flex1 and flex2 specify if the gadget can be resized by its parent, and how much.
 * When specified as a string, the format is:
 *     format = "flex1 flex2 flextype"  (flextype is optional)
 *
 * flex1 and flex2 have diffrent meanings depending on flextype:
 * - unspecified or "normal": flex1 means horizontal, flex2 vertical
 * - "self"      : flex1 is the main direction of the gadget, flex2 its cross direction
 * - "parent" : flex1 is the main direction of the gadget's parent, flex2 its cross direction.
 * Main/cross directions depend on the GFlow prop.
 *
 * flex1 and flex2 can being either:
 * - 0 (or fixed or f) : the gadget can't be resized by its parent in this direction
 * - 1 (or x), 2, 3, etc. : the gadget can be resized in this direction, and the value specify
 *  how much it can be resized compared to its siblings (2 means twice than 1, etc.)
 * The maximum value is MaxFlex (= 100)
 *
 * # Special cases:
 * - Inside a ScrollBox, 1 (or x), 2, 3, etc. mean that the size of gadget will grow when
 *  ScrollBox is resized. Hovever, it won't shrink (ie it will keep its maximum size).
 *
 * @note The actual  size also depends on the GSize prop of the gadget and on the GAlign prop of its parent.
 * @code
 *    auto& b1 = Button("Do It") << Size("20 auto") << Flex("1 1");
 *    auto& b2 = Button("Do It") << Size("20c a") << Flex("0 2");
 *    auto& b3 = Button("Do It") << Size("20c a") << Flex("0 2 parent");
 * @endcode
 * - b1: width = 20 pixels, height auto-computed, width and height can be changed by parent
 * - b2: width = 20 characters, height auto-computed, width can't be changed, height can be changed
 *   2 more times than siblings having a value of 1
 * - b3: same as b2 except that 0 and 2 refer to the main and cross directions of the parent
 */
class GFlex : public GVarProp_<GFlexValue> {
public:
  using Super = GVarProp_<GFlexValue>;
  enum {MinFlex = 0, MaxFlex = 100};
  enum Mode : int8_t {SelfWH, SelfFlow, ParentFlow};

  GFlex();
  GFlex(GString const& value);
  GFlex(int8_t flex1, int8_t flex2, Mode mode);
  GFlex(GFlex const&);
  GFlex(GFlexValue const&);

  GFlex* clone(bool copy_value) const override;

  GFlex& operator=(GFlex const& value) {set(value); return *this;}
  GFlex& operator=(GString const& value) {set(value); return *this;}

  using Super::set;
  void set(int8_t flex1, int8_t flex2, Mode);

  int8_t flex1() const {return value_.flex1;}
  int8_t flex2() const {return value_.flex2;}
  Mode mode() const {return Mode(value_.mode);}

  using GType = GPropType_<GFlex,GProp>;
  static GType& Type();
  GType& type() const override {return Type();}
  
  /// @internal
  /// @{
  bool onAdd(Gadget*) override;
  void onRemove(Gadget*) override;
  void updateGadget(Gadget&) override;
  void apply(Gadget*, GRender&, Specif) final;
  /// @}
};

/// Returns a GFlex created in the heap.
/// @{
inline GFlex& Flex() {return *new GFlex();}
inline GFlex& Flex(GString const& value) {return *new GFlex(value);}
/// @}

}
#endif
