//
//  Timers
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <chrono>
#include <gtimer.hpp>
#include <gtrigger.hpp>
#include <gcontrol.hpp>
#include <gnative.hpp>
#include <gstring.hpp>
using namespace std;
using namespace std::chrono;
namespace guit {

GIMPLEMENT_PROP(Timer, GPropType::Number)

GTimer& Timer(GTime duration, int count) {return *new GTimer(duration, count);}

GTimer::GTimer(GTime duration, int count) :
count_(count), duration_(duration) {}

GTimer::GTimer(GString const& arg) : GTimer(0, Infinite) {
  GStrings v;
  auto n = strsplit(arg, v, "");
  GString curloc = setlocale(LC_ALL,NULL);  // watch locale
  setlocale(LC_NUMERIC, "C");
  if (n > 0) setDelay(std::stoi(v[0].data()));
  if (n > 1) {
    auto count = strlower(v[1]);
    if (count == "infinite") setRepetitions(Infinite);
    else if (count == "oneshot") setRepetitions(OneShot);
    else count_ =  guit::toInt(count);
  }
  setlocale(LC_NUMERIC, curloc.data());
}

GTimer::~GTimer() {
  //cerr << "~GTimer "<< this << endl;
  if (running_) start(false);
  removeAll();
}

GTimer& GTimer::setDelay(GTime d) {
  duration_ = d;
  return *this;
}

GTimer& GTimer::setRepetitions(int c) {
  count_ = c;
  return *this;
}

void GTimer::start(bool yes) {
  if (running_) GCore::natApp()->stopTimer(*this);
  running_ = false;
  
  if (yes) {
    if (count_ != 0) {
      //set(0);   // set() calls a callback !!
      value_ = 0;
      // the timer cant be destroyed while it is running
      self_ = this;
      if (GCore::natApp()->startTimer(*this, duration_)) running_ = true;
    }
  }
}

void GTimer::restart() {
  if (!running_) return;
  GCore::natApp()->stopTimer(*this);
  if (GCore::natApp()->startTimer(*this, duration_)) running_ = true;
}

void GTimer::timerCB() {
  //cerr << "timerCB " << value_ << " waitingcalls: "<< waitingcalls_ << endl;
  // be sure the timer is not destroyed in a callback
  gprotect<GTimer> protect(this);

  if (count_ == Infinite) {   // repeeats indefinititely
    set(value_ + waitingcalls_, true);  // true => always call callbacks
    waitingcalls_ = 0;
  }
  //else if (val < count_ - 1) {   // not all repetitions done
  else if (value_ < count_) {
    set(value_ + waitingcalls_, true);  // true => always call callbacks
    waitingcalls_ = 0;
  }
  else {   // all repetitions done => stops or kills the timer
    waitingcalls_ = 0;
    start(false);
    GEvent e(GTrigger::done, GCore::elapsedTime());
    fireCallbacks(e);
    // kill the timer safely: adding and removing a reference will destroy
    // the timer only if there is no other ref pointing it
    if (count_ == OneShot) {gptr<GTimer> p = this;}
  }
  // the timer can be destroyed now
  if (!running_) self_ = nullptr;
}

GInt& GTimer::addBeat(GInt& prop, int tcount, int maxval) {
  prop.setRange(0, maxval-1);
  prop.setCycles(true);
  prop.bindImpl(*this,
                [this,&prop,tcount] {if (value() != 0 && value() % tcount == 0) prop++;},
                -1/*bindtag*/);
  return prop;
}

GInt& GTimer::addBeat(int tcount, int maxval) {
  auto& beat = *new GInt;
  return addBeat(beat, tcount, maxval);
}


}
