//
// Base class of variable Props (eg GInt, FText, GColor, GFont)
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <gfile.hpp>
#include <gvarprop.hpp>
#include <gevent.hpp>
#include <gexpr.hpp>
#include <gnotifier.hpp>
#include <gcontrol.hpp>
#include <gstyle.hpp>
#include <gadgetstyle.hpp>
using namespace std;
namespace guit {


GString GPropWrapper::stringValue() const {
  GString txt;
  gconvert(floatValue(), txt);
  return txt;
}

float GPropWrapper::floatValue() const {
  float val;
  gconvert(stringValue(), val);
  return val;
}


GPropType& GNotifier::Type() {
  static auto& t = *new GPropType("Notifier", GPropType::Notifier);
  return t;
}

GString GNotifier::stringValue() const {return "<Notifier>";}

GTriggerNotifier::GTriggerNotifier(GObject& sender, GExprTrigger& receiver)
: sender_(sender), recv_(receiver) {receiver.setNotifier(this);}

void GTriggerNotifier::fire(GEvent& e) {
  //if (e.matches(receiver_)) receiver_.fire(sender_, true);
  //else receiver_.fire(sender_, false);
  int val = e.matches(recv_, -1);
  if (val != 0) {
    if (val > 0) recv_.fire(sender_, true);
    else recv_.fire(sender_, false);
  }
}

void GTriggerNotifier::onRemove(Gadget* g)  {recv_.senderRemoved(sender_);}

GTriggerExprNotifier::GTriggerExprNotifier(GObject& sender, GExprTrigger& receiver, GBoolExpr& expr)
: GTriggerNotifier(sender, receiver), expr_(expr) {}

void GTriggerExprNotifier::fire(GEvent&) {recv_.fire(sender_, expr_.boolValue());}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

class GadgetNotifier : public GNotifier {
public:
  GadgetNotifier(Gadget& g) : gadget_(g){}
  void* receiver() override {return &gadget_;}
  void fire(GEvent&) override = 0;
  // normalement inutile car la prop appartient au gadget
  //void onRemove(gadget* g) override {gadget.removeNotifiers(&sender_);}
  Gadget& gadget_;
};


class GadgetUpdateNotifier : public GadgetNotifier {
public:
  GadgetUpdateNotifier(Gadget& g, GVarProp& prop) : GadgetNotifier(g) {}
  void fire(GEvent& e) override {e.fromProp().updateGadget(gadget_);}
};

class GadgetFunNotifier : public GadgetNotifier {
public:
  GadgetFunNotifier(Gadget& g, std::function<void(Gadget&)> const& fun)
  : GadgetNotifier(g), fun_(fun) {}
  void fire(GEvent&) override {if (fun_) (fun_)(gadget_);}
  std::function<void(Gadget&)> fun_;
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

class GChangeFunNotifier : public GChangeNotifier {
public:
  GChangeFunNotifier(GVarProp& sender, GVarProp& recv, GFunction const& fun)
  : sender_(sender), recv_(recv), fun_(fun) {}

  void* sender() override {return &sender_;}
  void* receiver() override {return &recv_;}
  GVarProp* receiverAsProp() override {return &recv_;}
  void fire(GEvent&) override {if (fun_) (fun_)();}  // match()?
  void onRemove(Gadget* g) override {recv_.removeNotifiers(&sender_);}
  void doSet() override {}
  bool doTest() override {return true;}

  GVarProp &sender_, &recv_;
  GFunction fun_;
};

class GChangedDestrNotifier : public GNotifier {
public:
  GChangedDestrNotifier(GVarProp& sender, GVarProp& recv,
                        GVarProp::BindTag bindtag)
  : bindtag_(bindtag), sender_(sender), recv_(recv) {}

  void fire(GEvent&) override {}
  void* receiver() override {return &recv_;}
  GVarProp* receiverAsProp() {return &recv_;}
  void onRemove(Gadget*) override {recv_.removeNotifiers(&sender_);}

  GVarProp::BindTag bindtag_;
  GVarProp &sender_, &recv_;
};

class GWrapDestrNotifier : public GNotifier {
public:
  GWrapDestrNotifier(GVarProp* sender, GProp* receiver)
  : sender_(sender), receiver_(receiver) {}

  void fire(GEvent&) override {}
  void* receiver() override {return receiver_;}
  void onRemove(Gadget*) override {receiver_->removeNotifiers(sender_);}

  GVarProp *sender_;
  GProp *receiver_;
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

//GVarProp::GVarProp() : modes_{} {}
//objmodes_.ismutable = 1; pas au bon endroit, ecrase const GObject() !

GVarProp::~GVarProp() {
  //cerr << " GGVarProp::~GGVarProp "<<this <<" " << endl;
  //if (omodes_.plain) {
  /* peut pas marcher, certains plain peuvent avoir ete detruits
   cerr << "~GGVarProp() plain  "<< this << " " << typeName() << endl;
   for (auto& it : props_) {
   if (auto* n = dynamic_cast<GadgetNotifier*>(it.get()))
   n->gadget_.remove(this);
   }
   */
  //}
  removeAll();
}

bool GVarProp::read(GString const& filename) {
  if (!isMutable()) {immutableError("read()"); return false;}
  GIFile in(filename);
  if (!in) return false;
  read(in);
  return true;
}

void GVarProp::read(std::istream& in) {
  if (!isMutable()) {immutableError("read()"); return;}
  GString str;
  std::getline(in, str);
  set(str);
}

void GVarProp::blend(GProp const& prop1, GProp const& prop2, float mix) {
  if (mix >= 0.5) set(prop1); else set(prop2);
}

void GVarProp::immutableError(const char* fun) {
  error(fun, "This "+ typeName() + " prop is immutable, its value can't be changed");
}

void GVarProp::propagateChange() {
  if (props_.empty()) return;
  // peut  entrainer des boucles oo, par ex si les ranges sont incompatibles
  auto* master = pmodes_.hasMaster ? getMaster() : nullptr;
  if (master) master->doSet();

  // avoids destructing object while firing callbacks
  gprotect<GVarProp> protect(this);
  GEvent e(GTrigger::changeaction, GCore::elapsedTime());
  e.aux_ = this;

  for (auto& it : props_.list_) {
    //cerr << "propagateChange " << this << " / " << it.get() << " " << master << endl;
    if (it.get() != master)
      it->fire(e);
  }
}

void GVarProp::fireCallbacks(GEvent& e) {
  if (props_.empty()) return;
  // avoids destructing object while firing callbacks
  gprotect<GVarProp> protect(this);
  e.aux_ = this;
  props_.fire(e);
}

void GVarProp::fireDefaultCallbacks() {
  GEvent e(GTrigger::changeaction, GCore::elapsedTime());
  fireCallbacks(e) ;
}

bool GVarProp::testMasterValue() {
  if (!pmodes_.hasMaster) return true;
  auto* master = getMaster();
  if (!master) return true;
  else return master->doTest();
}

GChangeNotifier* GVarProp::getMaster() {
  if (!isMutable() || !pmodes_.hasMaster) return nullptr;
  for (auto& it : props_) {
    auto* n = dynamic_cast<GChangeNotifier*>(it.get());
    if (n && n->recv_is_master_) return n;
  }
  return nullptr;
}

void GVarProp::bindImpl(GExpr& expr) {
  if (!isMutable()) {immutableError("bind()"); return;}
  auto* wrap = new GExprWrapper(expr);
  if (wrap->bindWith(*this)) {
    props_.addEnd(*wrap);
    add(*new GWrapDestrNotifier(this, wrap));
  }
  else delete wrap;
}

void GVarProp::bindImpl(GVarProp& sender, GFunction fun, BindTag tag) {
  if (!isMutable()) {immutableError("bind()"); return;}
  if (!sender.isMutable()) {sender.immutableError("bind()"); return;}

  auto& n = *new GChangeFunNotifier(sender, *this, fun);
  n.bindtag_ = tag;
  sender.add(n);
  add(*new GChangedDestrNotifier(*this, sender, tag));
  (fun)();  // inits value
}

void GVarProp::bindImpl(GVarProp& sender, GChangeNotifier& n,
                        BindTag tag, bool recv_is_master) {
  if (!isMutable()) {immutableError("bind()"); return;}
  if (!sender.isMutable()) {sender.immutableError("bind()"); return;}

  if (recv_is_master) {
    // faudrait aussi le faire dans l'autre sens si ils coincident !!!
    if (auto* m = sender.getMaster()) m->recv_is_master_ = false; // enlever precedent
    sender.pmodes_.hasMaster = 1;
    n.recv_is_master_ = true;
  }
  n.bindtag_ = tag;
  sender.add(n);
  add(*new GChangedDestrNotifier(*this, sender, tag));   // economiser ?
  if (!recv_is_master) n.doSet();  // init except if master
}

// NOTE: enleve both ways sans tenir compte des fields!
void GVarProp::unbind(GVarProp& sender) {
  if (!isMutable()) {immutableError("unbind()"); return;}
  if (!sender.isMutable()) {sender.immutableError("unbind()"); return;}

  if (pmodes_.hasMaster) {
    auto* m = getMaster();
    if (m && m->receiver() == &sender) pmodes_.hasMaster = 0;
  }
  if (sender.pmodes_.hasMaster) {
    auto* m = sender.getMaster();
    if (m && m->receiver() == this) sender.pmodes_.hasMaster = 0;
  }
  removeNotifiers(&sender);
  sender.removeNotifiers(this);
}

void GVarProp::unbind(BindTag tag) {
  // pas bon pour 2 raisons :
  // a) va enlever des bindings qu'il faut gader: expe int lie avec text de textfield
  // b) la boucle est invalidee qunad on fait unbind
  if (!isMutable()) {immutableError("unbind()"); return;}

  bool doagain = false;
  do {
    doagain = false;

    for (auto& it : props_) {
      if (auto* n = it->toNotifier()) {

        if (auto* n1 = dynamic_cast<GChangeNotifier*>(n)) {
          if (tag < 0 || tag == n1->bindtag_) {
            if (auto* recv = n1->receiverAsProp()) {
              unbind(*recv);
              doagain = true;
              break;
            }
          }
        }
        else if (auto* n2 = dynamic_cast<GChangedDestrNotifier*>(n)) {
          if (tag < 0 || tag == n2->bindtag_) {
            if (auto* recv = n2->receiverAsProp()) {
              recv->unbind(*this);
              doagain = true;
              break;
            }
          }
        }
      }
    }
  } while (doagain);
}

void GVarProp::addFun(GFunction const& fun) {
  if (!isMutable()) immutableError("add(function)");
  else props_.addEnd(on::changeaction / fun);
}

void GVarProp::add(GCond& c) {
  if (!isMutable()) immutableError("add(GCond)");
  else props_.addEnd(c);
}

void GVarProp::add(GProp& p) {
  if (!isMutable()) immutableError("add(GProp)");
  else props_.addEnd(p);
}

void GVarProp::addBegin(GProp& c) {
  if (!isMutable()) immutableError("addBegin(GProp)");
  else props_.addBegin(c);
}

void GVarProp::remove(GProp& c) {
  if (!isMutable()) immutableError("remove(GProp)");
  else props_.remove(nullptr, c);
}

void GVarProp::removeAll() {
  if (!isMutable()) return;  // cant remove
  if (pmodes_.removing) return;  // to avoid infinite loops because of crossed refs
  pmodes_.removing = 1;
  props_.removeAll(nullptr);
  pmodes_.removing = 0;
}

void GVarProp::update() {
  for (auto& it : props_) {
    if (auto* n = dynamic_cast<GadgetNotifier*>(it.get())) n->gadget_.update();
  }
}

void GVarProp::perform(std::function<void(Gadget&)> fun) {
  for (auto& it : props_) {
    if (auto* n = dynamic_cast<GadgetNotifier*>(it.get()))
      (fun)(n->gadget_);
  }
}

void GVarProp::removeFromGadgets() {
  gprotect<GVarProp>  protect(this);
  while (!props_.empty()) {    // caution remove() changes prop list!
    for (auto& it : props_) {
      if (auto* n = it->toNotifier()) {
        if (auto* gn = dynamic_cast<GadgetNotifier*>(n)) {
          //cerr << "remove from " << gn << endl;
          gn->gadget_.remove(this);
          break;
        }
      }
    }
  }
}

// reste-t'il au moins un notifier non GadgetNotifier?
bool GVarProp::hasActiveNotifiers() {
  for (auto& it : props_) {
    if (auto* n = it->toNotifier()) {
      if (!dynamic_cast<GadgetNotifier*>(n)) return true;
    }
  }
  return false;
}

void GVarProp::removeNotifiers(void* obj) {
  if (!obj) return;
  if (!isMutable()) {immutableError("removeNotifiers()"); return;}
  if (pmodes_.removing) return;  // to avoid infinite loops because of crossed refs
  pmodes_.removing = 1;
  props_.removeNotifiers(obj);
  pmodes_.removing = 0;
}

void GVarProp::startNotifyGadget(Gadget& g) {
  if (!isMutable()) {immutableError("startNotify()"); return;}
  add(*new GadgetUpdateNotifier(g, *this));
  updateGadget(g);
}

void GVarProp::startNotify(Gadget& g, std::function<void(Gadget&)> const&fun) {
  if (!isMutable()) {immutableError("startNotify()"); return;}
  add(*new GadgetFunNotifier(g, fun));
}

void GVarProp::stopNotify(Gadget& g) {
  if (!isMutable()) {immutableError("endNotify()"); return;}
  removeNotifiers(&g);
}

}
