//
//  Misc Gadgets
//  guit GUI Toolkit
//  Copyright © 2019/2020 Eric Lecolinet. All rights reserved.
//  http://www.telecom-paris.fr/~elc
//

#include <clocale>
#include <map>
#include <unistd.h>
#include <list>
#include <iostream>
#include <guit.hpp>
using namespace std;
using namespace guit;

class GUI {
  GBox& menubar = MenuBar();
  GBox& toolbar = ToolBar();
  GBox& statusbar = StatusBar();
  GText& status = Text();
  
  void doIt() {
    std::cout << "DoIt "<< std::endl;
  }
  
  void echo(const std::string& msg) {
    status = msg;
    std::cout << msg << std::endl;
  }

  void openDoc(const std::string& filename) {
    echo("Opening file: " + filename);
  }
  
  void saveDoc(const std::string& filename) {
    echo("Saving file: " + filename);
  }
  
  void testThread() {
    for (int i = 0; i < 10; ++i) {
      cerr << "testThread " << i <<endl;
      gexec([this,i]{echo("testThread " + to_string(i));});
      sleep(1);
    }
  }
  
  static void testThread2() {
    for (int i = 0; i < 10; ++i) {
      cerr << "testThread2 " << i <<endl;
      sleep(1);
    }
  }
  
  static void testMain() {
    for (int i = 0; i < 10; ++i) {
      cerr << "testMain " << i <<endl;
      sleep(1);
    }
  }

  auto& MyRadiobox(GString const& title) {return RadioBox(".MyRadiobox "+title);}

public:

  GUI() {

    auto& top_part_shown = Check(true);
    auto& bottom_part_shown = Check(true);
    auto& colorchoice = Choice();

    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    // Dialog boxes
    
    // Open Dialog
    auto& opendial = OpenDialog();
    opendial << [&]{openDoc(opendial.pathname());};
    
    // Save Dialog
    auto& savedial = SaveDialog();
    savedial << [&]{saveDoc(savedial.pathname());};
    
    // Quit Dialog
    auto& quitdial = QuitDialog();

    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    // MEBUBAR
    auto& filemenu = Menu("#file-menu")
        << ("{icon:open.png; hotkey:ctrl-O} Open..." << opendial)
        << ("{icon:save.png; hotkey:ctrl-S} Save..." << savedial)
        << Separator
        << ("#quit" << quitdial);
    
    menubar
    << ("File" << filemenu)
    << ("View"
        << (Menu
            << (CheckItem
                << on::checked   / "Show Toolbar"
                << on::unchecked / "Hide Toolbar"
                << [this]{toolbar.show(!toolbar.isShown());}
                )
            << (CheckItem
                << on::checked   / "Hide Top Part"
                << on::unchecked / "Show Top Part"
                << top_part_shown
                )
            << (CheckItem
                << on::checked   / "Hide Bottom Part"
                << on::unchecked / "Show Bottom Part"
                << bottom_part_shown
                )
            << (CheckItem
                << on::checked   / "Show Statusbar"
                << on::unchecked / "Hide Statusbar"
                << [this]{statusbar.show(!statusbar.isShown());}
                )
            )
        )
    << ("Colors"
        << (Menu()
            << colorchoice
            << RadioItem("No Color")
            << RadioItem("Yellow")
            << RadioItem("Red")
            )
        )
    << ("Help"
        << (Menu
            << (HBox
                << "    Search"s
                << TextField("{Size: 100 auto}")
                << Button("Ok")
                )
            << ("Documentation" << [this]{echo("Not Yet Available");})
            << ("Version" << [this]{echo("Version 1.0");})
            << ("Author"  << [this]{echo("Eric Lecolinet - Telecom ParisTech");})
            )
        );
    
    
    toolbar
    << ("{icon:open.png; hotkey:ctrl-O} Open" << opendial)
    << ("Save" << savedial)
    << ("Quit" << quitdial)
    << ("Clear Status" << [this]{status = "";})
    << (Item("Action")
        << Tip("Action")
        << [](auto&e){cerr << "click: "<< e.asMouseEvent().clickCount() <<endl;}
        )
    << (Item("{tip: Click} Click")
        << on::click    / []{cerr << "click"<<endl;}
        << on::dblclick / []{cerr << "dblclick"<<endl;}
        )
    ;

    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    // TOP PART

    auto& dir1 = Flow();
    auto& dir2 = Flow();
    dir1 = GFlow::row;
    dir2 = GFlow::column;
    
    auto& bgcolor = Background("none");
    auto& fruits = ListBox("#fruits");
        
    auto& top_part = VBox()
    << (Show() <<= top_part_shown)
    << dir1
    << (RadioBox
        << dir2
        << bgcolor
        << (Button("Vertical")
            << Check(true)
            << [&]{dir1 = GFlow::row; dir2 = GFlow::column;} // changes the direction
            )
        << (Button("Horizontal")
            << [&]{dir1 = GFlow::column; dir2 = GFlow::row;}
            )
        /* pas implementé
        << (Button("Test1")
            << on::mousedown / (dir1 << Dir.column)
            )
        << (Button("Test2")
            << on::action / (Dialog << "aaa" << "bbb")
            )
        */
        )
    << (RadioBox(".colors")
        << dir2
        << bgcolor
        << colorchoice
        << ("No Color" << [&]{bgcolor = GBackground::none;})
        << ("Yellow"   << [&]{bgcolor = GBackground::yellow;})
        << ("Red"      << [&]{bgcolor = GBackground::red;})
        )
    << (fruits
        << dir2
        << bgcolor
        << "Apples"
        << ("Oranges" << on::checked / GBackground::orange)
        << "Cherries"
        << "Bananas"
        )
    << (ButtonBox("{border: rounded}")
        << dir2
        << bgcolor
        << (CheckItem
            << Check(true)
            << on::checked   / "Check 1" / GColor::orange
            << on::unchecked / "Check 1" / GColor::gray
            )
        <<  (CheckItem("Check 2") << []{cout << "check2\n";})
        <<  (CheckItem("Check 3") << []{cout << "check3\n";})
        );
    
    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    // MIDDLE PART
    
    auto& color_name = Text();
    auto& fruit_name = Text();
    
    auto& result = Text();
    result <<= color_name;
    result <<= fruit_name;
    
    auto& table = Table("{border: rounded; size: a a; flex: 1 0}")
    << (TRow
        << "Color:"s
        << (TextField
            << color_name
            << (~colorchoice == 1) / ("Yellow - index: " + ~colorchoice)
            << (~colorchoice == 2) / ("Red - index: " + ~colorchoice)
            << (~colorchoice == 3) / ("No Color - index: " + ~colorchoice)
            )
        << (Button("Result") << [&]{result = color_name();})
        )
    << (TRow
        << "Fruit:"s
        << Tip("Fruit")
        << (TextField
            << fruit_name
            << (~fruits.choice() == 0) / ("Apples - index: " + ~fruits.choice())
            << (~fruits.choice() == 1) / ("Oranges - index: " + ~fruits.choice())
            << (~fruits.choice() == 2) / ("Cherries - index: " + ~fruits.choice())
            << (~fruits.choice() == 3) / ("Bananas - index: " + ~fruits.choice())
            )
        << (Button("Result") << [&]{result = fruit_name();})
        )
    << (TRow
        << "Result:"s
        << (TextField << result)
        << (Button("Clear") << [&]{result = "";})
        );

    //auto& txt = TextArea("aaaa toto tata titi \nbbb toto tata titi \ncccc xxxx yyyy")

    auto& txt = TextArea("aaa aaaa aaaa aaaa\nbb bb  bbbbb\nggg ggggg  gg\ncc cccc cc")    // MEGA BUG !!!!!
    << new GSize("100 a") << Align("c c");
    txt.setMultiLine(true);

    auto& btn = Button("aaaa\nbbbbb\ncccc")
    << new GSize("100 a") << Align("c c");
    btn.setMultiLine(true);

    auto& middle_part = HBox()
    << table
    << txt
    << btn;
    //truc = &middle_part;

    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    // BOTTOM PART

    auto& tf_color = Color();
    auto& tf_font = Font();
    auto& img_just = Align();

    auto& slider_value = Float();
    slider_value.setRange(-100, 100);
                       
    auto& attributes = Table("{border: rounded; size: a a; flex: 1 0}")
    << (TRow
        << ""s
        << (HBox
            << (TextField("abééïïcdexyzw") << tf_color << tf_font)
            // << (Button("{size: a a 0 0} C") << [&]{text << "";})
            )
        )
    << (TRow
        << "Attributes:"s
        << (HBox("{margin: 0 0}")
            << (OptionBox
                << [&](GEvent& e){tf_color = e.from().text();}
                << Label("<i>Color")
                << "Black" << "Red" << "Blue" << "Green"
                << Choice(0)
                )
            << (OptionBox
                << [&](GEvent& e){tf_font = e.from().text();}
                << Label("<i>Font")
                << "Regular" << "Bold" << "Italic" << "Bold Italic"
                << Choice(0)
                )
            )
        )
    << (TRow
        << ""s
        << (Button("{icon: icon.png} Image") << img_just)
        )
    << (TRow
        << "Justify:"s
        << (HBox
            << (OptionBox
                << Label("<i>Text")
                << ("Start"  << [&]{img_just.setMainLayout(GAlign::Start);})
                << ("Center" << [&]{img_just.setMainLayout(GAlign::Center);})
                << ("End"    << [&]{img_just.setMainLayout(GAlign::End);})
                << Choice(0)
                )
            << (OptionBox
                << Label("<i>Icon")
                << ("Start"  << [&]{img_just.setIconLayout(GAlign::Start);})
                << ("Center" << [&]{img_just.setIconLayout(GAlign::Center);})
                << ("End"    << [&]{img_just.setIconLayout(GAlign::End);})
                << Choice(0)
                )
            )
        )
    << (TRow
        << "Value:"s
        << (HBox
            // << Margin(0,0)
            << Margin()
            << (TextField("{size: 5C auto; flex: 0 1}") << slider_value)
            << (Slider("{size: 100 a; flex: 1 1}") << slider_value)
            )
        );
    
    auto& tm = Text();
    auto& tx = Text("red");
    auto& col = Color();
    auto& col2 = Color("red");
    auto& lab = Label("label");
    
    auto& test = Table()
    << (TRow
        << "Label:"s
        << (HBox
            << Align("start start")
            << (TextField << *new GSize("a a") << "xxxx  ")
            << (TextField << tm << "yyyy")
            << (Label << tm)
            << "Toto"
            )
        )
    << (TRow
        << (Label("text") << (col <<= tx))
        << (TextField << tx)
        )
    << (TRow
        << (lab << col2)
        << (HBox()
            << (Button("add") << [&]{lab << GColor::blue ;})
            << (Button("set") << [&]{col2 = GColor::green ;})
            )
        )
    << (TRow
        << (Button
            << on::mousedown / "<color=red>text when down"
            << on::mouseup   / "<color=blue>text when up"
            )
        << (Button
            << "<color=blue>text when down"s
            << on::mousedown / "<color=red>text when up"
            )
        );

    
    auto& scrollbox = ScrollBox("{size: a 8C; background: white")
    << (RadioBox("{align:s x}")
        << "one"
        << "two"
        << "three"
        << (RadioBox("{flow:row}")
            << Button("{check:true} 1 ")
            << Button(" 2 ")
            << Button(" 3 ")
            << Button(" 4 ")
            )
        << TextField("{check:true} aaa")
        << TextField("bbb")
        << TextField("ccc")
        << TextField("ddd")
        << "four"
        << "five"
        << "six"
        << "seven"
        << "eigth"
        << "nine"
        << "ten"
        << "eleven"
        << "twelve"
        );
    
    auto& bottom_part = HBox()
    << (Show() <<= bottom_part_shown)
    << (VBox("{size: a a; flex: 0 1}")
        << attributes
        << test
        )
    << scrollbox;

    auto& window = Window("#gui")
    << menubar
    << toolbar
    << top_part
    << middle_part
    << (bottom_part << new GSize("a a") << new GFlex("1 5"))
    << (statusbar << (Label << status));

    echo("Ready");
    
    //Color.red = Color.blue;  FAUX car immutable!

    window.startApp();  // starts the app => blocking, must be last
  }
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

int main(int argc, const char *argv[]) {
  GApp app(argc, argv);
  app << "#gui {icon: icon.png; text: Misc. Gadgets}"
  "#file-menu #quit {icon:quit.png; hotkey:ctrl-Q; text: Quit; color:red; font:bold}"
  ".colors {value: No Color}"
  "#fruits {choice: 1}";

  GUI gui;
  return 0;
}
